# encoding: utf-8
'''
srfnef.utils
~~~~~~~~~~~~

This module provides utility functions that are used within SRF-NEF
that are alose useful for extenel comsumptions.
'''

import getpass
import os
import platform
import sys
import time

import numpy as np
import tqdm
from matplotlib import pyplot as plt


def is_notebook():
    '''check if the current environment is `ipython`/ `notebook`
    '''
    return 'ipykernel' in sys.modules


is_ipython = is_notebook


def _tqdm(*args, **kwargs):
    '''same as tqdm.tqdm
    Automatically switch between `tqdm.tqdm` and `tqdm.tqdm_notebook` accoding to the runtime
    environment.
    '''
    if is_notebook():
        return tqdm.tqdm_notebook(*args, **kwargs)
    else:
        return tqdm.tqdm(*args, **kwargs)


def _dataclass_to_md(dat, indent = 0):
    out_string = ''
    out_string += '   ' * indent + f'- {dat.__class__.__name__}:\n\n'
    if callable(dat):
        out_string += '   ' * (indent + 1) + f'- callable\n\n'
    for key, value in dat.to_dict(recurse = False).items():
        from srfnef.templating import DataClass
        if isinstance(value, DataClass):
            out_string += _dataclass_to_md(value, indent + 1)
        elif key == 'data':
            out_string += '   ' * (
                    indent + 1) + f'- {key}: array(shape = {value.shape}, dtype = {value.dtype})\n\n'
        else:
            out_string += '   ' * (indent + 1) + f'- {key}: {value}\n\n'
    return out_string + '\n'


def _imshow(dat):
    out = ''
    if isinstance(dat, list):
        for img in dat:
            out += _dataclass_to_md(img)
        n = len(dat)

        plt.figure(figsize = (3 * n, 3))
        for i in range(n):
            plt.subplot(1, n, i + 1)
            dat[i].imshow()
        else:
            plt.savefig(f'recon_image.png')

        out += f'![alt text](recon_image.png)\n\n'

        plt.figure(figsize = (3 * n, 3))
        for i in range(n):
            p0, p1 = dat[i].central_profiles[:2]
            shape, center, size, unit_size = dat[i].shape, dat[i].center, dat[i].size, dat[
                i].unit_size
            coo_0 = (np.arange(shape[0]) + 0.5) * unit_size[0] - size[0] / 2 + center[0]
            coo_1 = (np.arange(shape[1]) + 0.5) * unit_size[1] - size[1] / 2 + center[1]
            plt.subplot(1, n, i + 1)
            plt.plot(coo_0, p0, label = 'x profile')
            plt.plot(coo_1, p1, label = 'y profile')
        else:
            plt.savefig(f'recon_profile.png')
        out += f'![alt text](recon_profile.png)'

    else:
        i = 0
        plt.figure(figsize = (3, 3))
        dat.imshow()
        plt.savefig(f'recon_image_{i}.png')

        plt.figure(figsize = (3, 3))
        plt.subplot(211)
        p0, p1 = dat.central_profiles[:2]
        shape, center, size, unit_size = dat.shape, dat.center, dat.size, dat.unit_size
        coo_0 = (np.arange(shape[0]) + 0.5) * unit_size[0] - size[0] / 2 + center[0]
        coo_1 = (np.arange(shape[1]) + 0.5) * unit_size[1] - size[1] / 2 + center[1]
        plt.plot(coo_0, p0, label = 'x profile')
        plt.plot(coo_1, p1, label = 'y profile')
        plt.savefig(f'recon_profile_{i}.png')
        out += f'![alt text](recon_image_{i}.png)![alt text](recon_profile_{i}.png)\n\n'
    return out


def write_markdown(path, args, output):
    out_string = ''
    out_string += f'# Title: test generate markdown doc \n\n'
    out_string += f'Time: {time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}\n\n'
    out_string += f'Author: {getpass.getuser()}\n\n'
    for arg in args:
        out_string += '---\n\n'
        out_string += _dataclass_to_md(arg)
    out_string += '---\n\n'
    out_string += '## Output: \n\n'
    out_string += _imshow(output)
    with open(path, 'w', encoding = 'utf8') as fout:
        fout.write(out_string)


__all__ = (
    '_eps', '_small', '_tiny', '_huge', '_pi', 'resource_path', 'separator')

_eps = 1e-8

_small = 1e-4

_tiny = 1e-8

_huge = 1e8

_pi = 3.14159265358979323846264338

if 'Windows' in platform.system():
    separator = '\\'
else:
    separator = '/'

resource_path = os.path.abspath(
    os.path.dirname(os.path.abspath(__file__)) + (
            separator + '..') * 1 + separator + 'resources' + separator)

toy_config_path = resource_path + 'toy_config.json'
toy_emap_path = resource_path + 'emap' + separator + 'sum_emap.npy'



from copy import deepcopy

import attr
import numpy as np
import scipy.ndimage as nd
from numba import jit
from scipy import sparse

from .templating import dataclass, TYPE_BIND
from .utils import is_notebook, _tqdm


@dataclass
class Block:
    size: np.ndarray = attr.ib(converter = np.array)
    shape: np.ndarray = attr.ib(converter = np.array)

    @property
    def unit_size(self):
        return np.array([self.size[i] / self.shape[i] for i in range(len(self.size))])


@dataclass
class PETCylindricalScanner:
    inner_radius: float
    outer_radius: float
    nb_rings: int
    nb_blocks_per_ring: int
    gap: float
    blocks: Block

    @property
    def axial_length(self):
        return self.blocks.size[2] * self.nb_rings + self.gap * (self.nb_rings - 1)

    @property
    def central_bin_size(self):
        return 2 * np.pi * self.inner_radius / self.nb_detectors_per_ring / 2

    @property
    def nb_detectors_per_block(self):
        return np.prod(self.blocks.shape)

    @property
    def nb_detectors_per_ring(self):
        return self.nb_detectors_per_block * self.nb_blocks_per_ring

    @property
    def nb_detectors(self):
        return self.nb_detectors_per_ring * self.nb_rings

    @property
    def angle_per_block(self):
        return 2 * np.pi / self.nb_blocks_per_ring

    @property
    def nb_thin_rings(self):
        return self.nb_rings * self.blocks.shape[2]

    @property
    def nb_detectors_per_thin_ring(self):
        return self.blocks.shape[1] * self.nb_blocks_per_ring


@dataclass
class LORs:
    data: np.ndarray

    @property
    def fst(self):
        return self.data[:, :3]

    @property
    def snd(self):
        return self.data[:, 3:]

    @property
    def shape(self):
        return self.data.shape

    def __len__(self):
        return self.data.shape[0]

    def __bool__(self):
        return self.__len__() > 0

    @classmethod
    def from_scanner(cls, scanner):
        _lors = _mesh_detector(scanner)
        return cls(_lors)

    @classmethod
    def from_scanner_ring(cls, scanner, i, j):
        _lors = _mesh_detector_ring(scanner, i, j)
        return cls(_lors)

    @classmethod
    def from_scanner_thin_ring(cls, scanner, i, j):
        _lors = _mesh_detector_thin_ring(scanner, i, j)
        return cls(_lors)

    @classmethod
    def from_fst_snd(cls, fst, snd):
        return cls(np.hstack((fst, snd)))


#
# def _mesh_detector(scanner):
#     lors = np.zeros((scanner.nb_detectors * scanner.nb_detectors, 6), dtype = np.float32)
#
#     R = (scanner.inner_radius + scanner.outer_radius) / 2
#     theta = 2 * np.pi / scanner.nb_blocks_per_ring * np.arange(scanner.nb_blocks_per_ring)
#     x = np.ones(scanner.blocks.shape[1], ) * R
#     y = (np.arange(scanner.blocks.shape[1]) + 0.5) * scanner.blocks.unit_size[1] - \
#         scanner.blocks.size[1] / 2
#     z = (np.arange(scanner.blocks.shape[2]) + 0.5) * scanner.blocks.unit_size[2]
#     xx = np.kron(x, [1] * scanner.nb_blocks_per_ring)
#     yy = np.kron(y, [1] * scanner.nb_blocks_per_ring)
#     theta1 = np.kron(theta, [[1]] * scanner.blocks.shape[1]).ravel()
#     xx1 = xx * np.cos(theta1) - yy * np.sin(theta1)
#     yy1 = xx * np.sin(theta1) + yy * np.cos(theta1)
#
#     xx2 = np.kron(xx1, [[1]] * scanner.blocks.shape[2]).ravel()
#     yy2 = np.kron(yy1, [[1]] * scanner.blocks.shape[2]).ravel()
#     zz2 = np.kron(z, [1] * scanner.blocks.shape[1] * scanner.nb_blocks_per_ring).ravel()
#
#     xd = np.kron(xx2, [[1]] * scanner.nb_rings).ravel()
#     yd = np.kron(yy2, [[1]] * scanner.nb_rings).ravel()
#     zd = np.kron(zz2, [[1]] * scanner.nb_rings).ravel()
#     for i in range(scanner.nb_rings):
#         zd[i * scanner.nb_detectors_per_ring + np.arange(scanner.nb_detectors_per_ring)] += \
#             i * scanner.blocks.size[2] + i * scanner.gap - scanner.axial_length / 2
#
#     lors[:, 0] = np.kron(xd, [1] * scanner.nb_detectors)
#     lors[:, 1] = np.kron(yd, [1] * scanner.nb_detectors)
#     lors[:, 2] = np.kron(zd, [1] * scanner.nb_detectors)
#     lors[:, 3] = np.kron(xd, [[1]] * scanner.nb_detectors).ravel()
#     lors[:, 4] = np.kron(yd, [[1]] * scanner.nb_detectors).ravel()
#     lors[:, 5] = np.kron(zd, [[1]] * scanner.nb_detectors).ravel()
#
#     return lors


def _mesh_detector(scanner):
    lors = np.zeros((scanner.nb_detectors * scanner.nb_detectors, 6), dtype = np.float32)

    R = (scanner.inner_radius + scanner.outer_radius) / 2
    theta = 2 * np.pi / scanner.nb_blocks_per_ring * np.arange(scanner.nb_blocks_per_ring)
    x = np.ones(scanner.blocks.shape[1], ) * R
    y = (np.arange(scanner.blocks.shape[1]) + 0.5) * scanner.blocks.unit_size[1] - \
        scanner.blocks.size[1] / 2
    z = (np.arange(scanner.blocks.shape[2]) + 0.5) * scanner.blocks.unit_size[2]
    xx = np.kron(x, [1] * scanner.nb_blocks_per_ring)
    yy = np.kron(y, [1] * scanner.nb_blocks_per_ring)
    theta1 = np.kron(theta, [[1]] * scanner.blocks.shape[1]).ravel()
    xx1 = xx * np.cos(theta1) - yy * np.sin(theta1)
    yy1 = xx * np.sin(theta1) + yy * np.cos(theta1)

    xx2 = np.kron(xx1, [[1]] * scanner.blocks.shape[2]).ravel()
    yy2 = np.kron(yy1, [[1]] * scanner.blocks.shape[2]).ravel()
    zz2 = np.kron(z, [1] * scanner.blocks.shape[1] * scanner.nb_blocks_per_ring).ravel()

    xd = np.kron(xx2, [[1]] * scanner.nb_rings).ravel()
    yd = np.kron(yy2, [[1]] * scanner.nb_rings).ravel()
    zd = np.kron(zz2, [[1]] * scanner.nb_rings).ravel()
    for i in range(scanner.nb_rings):
        zd[i * scanner.nb_detectors_per_ring + np.arange(scanner.nb_detectors_per_ring)] += \
            i * scanner.blocks.size[2] + i * scanner.gap - scanner.axial_length / 2

    lors[:, 0] = np.kron(xd, [1] * scanner.nb_detectors)
    lors[:, 1] = np.kron(yd, [1] * scanner.nb_detectors)
    lors[:, 2] = np.kron(zd, [1] * scanner.nb_detectors)
    lors[:, 3] = np.kron(xd, [[1]] * scanner.nb_detectors).ravel()
    lors[:, 4] = np.kron(yd, [[1]] * scanner.nb_detectors).ravel()
    lors[:, 5] = np.kron(zd, [[1]] * scanner.nb_detectors).ravel()

    return lors



def _mesh_detector_ring(scanner, i, j):
    lors = np.zeros((scanner.nb_detectors_per_ring * scanner.nb_detectors_per_ring, 6),
                    dtype = np.float32)

    R = (scanner.inner_radius + scanner.outer_radius) / 2
    theta = 2 * np.pi / scanner.nb_blocks_per_ring * np.arange(scanner.nb_blocks_per_ring)
    x = np.ones(scanner.blocks.shape[1], ) * R
    y = (np.arange(scanner.blocks.shape[1]) + 0.5) * scanner.blocks.unit_size[1] - \
        scanner.blocks.size[1] / 2
    z = (np.arange(scanner.blocks.shape[2]) + 0.5) * scanner.blocks.unit_size[2]
    xx = np.kron(x, [1] * scanner.nb_blocks_per_ring)
    yy = np.kron(y, [1] * scanner.nb_blocks_per_ring)
    theta1 = np.kron(theta, [[1]] * scanner.blocks.shape[1]).ravel()
    xx1 = xx * np.cos(theta1) - yy * np.sin(theta1)
    yy1 = xx * np.sin(theta1) + yy * np.cos(theta1)

    xx2 = np.kron(xx1, [[1]] * scanner.blocks.shape[2]).ravel()
    yy2 = np.kron(yy1, [[1]] * scanner.blocks.shape[2]).ravel()
    zz2 = np.kron(z, [1] * scanner.blocks.shape[1] * scanner.nb_blocks_per_ring).ravel()

    xd = xx2
    yd = yy2
    zd = zz2

    lors[:, 0] = np.kron(xd, [1] * scanner.nb_detectors_per_ring)
    lors[:, 1] = np.kron(yd, [1] * scanner.nb_detectors_per_ring)
    lors[:, 2] = np.kron(zd + i * scanner.blocks.size[2] + i * scanner.gap,
                         [1] * scanner.nb_detectors_per_ring)
    lors[:, 3] = np.kron(xd, [[1]] * scanner.nb_detectors_per_ring).ravel()
    lors[:, 4] = np.kron(yd, [[1]] * scanner.nb_detectors_per_ring).ravel()
    lors[:, 5] = np.kron(zd + j * scanner.blocks.size[2] + j * scanner.gap,
                         [[1]] * scanner.nb_detectors_per_ring).ravel()

    return lors


def _mesh_detector_thin_ring(scanner, i, j):
    lors = np.zeros((scanner.nb_detectors_per_thin_ring * scanner.nb_detectors_per_thin_ring, 6),
                    dtype = np.float32)

    R = (scanner.inner_radius + scanner.outer_radius) / 2
    theta = 2 * np.pi / scanner.nb_blocks_per_ring * np.arange(scanner.nb_blocks_per_ring)
    x = np.ones(scanner.blocks.shape[1], ) * R
    y = (np.arange(scanner.blocks.shape[1]) + 0.5) * scanner.blocks.unit_size[1] - \
        scanner.blocks.size[1] / 2
    z = 0.5 * scanner.blocks.unit_size[2]
    xx = np.kron(x, [1] * scanner.nb_blocks_per_ring)
    yy = np.kron(y, [1] * scanner.nb_blocks_per_ring)
    theta1 = np.kron(theta, [[1]] * scanner.blocks.shape[1]).ravel()
    xx1 = xx * np.cos(theta1) - yy * np.sin(theta1)
    yy1 = xx * np.sin(theta1) + yy * np.cos(theta1)

    xd = np.kron(xx1, [[1]] * 1).ravel()
    yd = np.kron(yy1, [[1]] * 1).ravel()
    zd = np.kron(z, [1] * scanner.blocks.shape[1] * scanner.nb_blocks_per_ring).ravel()

    lors[:, 0] = np.kron(xd, [1] * scanner.nb_detectors_per_thin_ring)
    lors[:, 1] = np.kron(yd, [1] * scanner.nb_detectors_per_thin_ring)
    lors[:, 2] = np.kron(zd + i * scanner.blocks.unit_size[2] + i * scanner.gap,
                         [1] * scanner.nb_detectors_per_thin_ring)
    lors[:, 3] = np.kron(xd, [[1]] * scanner.nb_detectors_per_thin_ring).ravel()
    lors[:, 4] = np.kron(yd, [[1]] * scanner.nb_detectors_per_thin_ring).ravel()
    lors[:, 5] = np.kron(zd + j * scanner.blocks.unit_size[2] + j * scanner.gap,
                         [[1]] * scanner.nb_detectors_per_thin_ring).ravel()

    return lors


@dataclass
class Sinogram:
    data: sparse.csr_matrix
    scanner: PETCylindricalScanner

    @property
    def shape(self):
        return self.data.shape

    @property
    def vproj(self):
        return self.data

    @classmethod
    def initializer(cls, scanner):
        D = np.ones((scanner.nb_detectors, scanner.nb_detectors), dtype = np.float32)
        return cls(sparse.csr_matrix(D), scanner)

    def to_listmode(self):
        return _sinogram_to_listmode(self)


def _sinogram_to_listmode(sino: Sinogram):
    scanner = sino.scanner
    csr = sino.data
    row, col = csr.nonzero()
    iy1 = row % scanner.blocks.shape[1]
    ib1 = (row // scanner.blocks.shape[1]) % scanner.nb_blocks_per_ring
    i_thin_ring1 = row // scanner.nb_detectors_per_thin_ring
    lors_data = np.zeros((csr.nnz, 6), dtype = np.float32)
    x0 = (scanner.inner_radius + scanner.outer_radius) / 2
    y0 = (iy1 + 0.5) * scanner.blocks.unit_size[1] - scanner.blocks.size[1] / 2
    theta = scanner.angle_per_block * ib1
    lors_data[:, 0] = x0 * np.cos(theta) - y0 * np.sin(theta)
    lors_data[:, 1] = x0 * np.sin(theta) + y0 * np.cos(theta)
    lors_data[:, 2] = (i_thin_ring1 + 0.5) * scanner.blocks.unit_size[2] - scanner.axial_length / 2

    iy2 = col % scanner.blocks.shape[1]
    ib2 = (col // scanner.blocks.shape[1]) % scanner.nb_blocks_per_ring
    i_thin_ring2 = col // scanner.nb_detectors_per_thin_ring
    #     print(iy1, iy2, ib1, ib2, i_thin_ring1, i_thin_ring2)
    x0 = (scanner.inner_radius + scanner.outer_radius) / 2
    y0 = (iy2 + 0.5) * scanner.blocks.unit_size[1] - scanner.blocks.size[1] / 2
    theta = scanner.angle_per_block * ib2
    lors_data[:, 3] = x0 * np.cos(theta) - y0 * np.sin(theta)
    lors_data[:, 4] = x0 * np.sin(theta) + y0 * np.cos(theta)
    lors_data[:, 5] = (i_thin_ring2 + 0.5) * scanner.blocks.unit_size[2] - scanner.axial_length / 2

    return Listmode(csr.data, LORs(lors_data))


@dataclass
class Image:
    """
    Image data with center and size info.
    """

    data: np.ndarray
    center: np.ndarray = attr.ib(converter = np.array)
    size: np.ndarray = attr.ib(converter = np.array)

    # quaternion: tuple

    @property
    def ndim(self):
        return len(self.data.shape)

    @property
    def unit_size(self):
        return np.array([self.size[i] / self.shape[i] for i in range(self.ndim)])

    def shift(self, dist):
        '''shift image in mm'''
        _dist = dist / self.unit_size
        return self.replace(data = nd.shift(self.data, _dist, order = 1))

    def rotate(self, angle):
        '''rotate x-y'''
        if angle == 0:
            return self
        return self.replace(data = nd.rotate(self.data, np.rad2deg(angle), order = 1,
                                             reshape = False))

    def zoom(self, scale):
        return self.replace(data = nd.zoom(self.data, scale, order = 1))

    def adder(self, o2, angle, dist):
        o2 = o2.rotate(angle)
        o2 = o2.shift(dist)
        return self.replace(data = self.data + o2.data)

    @property
    def central_slices(self):
        t0 = self.data[int(self.shape[0] / 2), :, :]
        t1 = self.data[:, int(self.shape[1] / 2), :]
        t2 = self.data[:, :, int(self.shape[2] / 2)]
        return t0, t1, t2

    @property
    def central_profiles(self):
        p0 = self.data[:, int(self.shape[1] / 2), int(self.shape[2] / 2)]
        p1 = self.data[int(self.shape[0] / 2), :, int(self.shape[2] / 2)]
        p2 = self.data[int(self.shape[0] / 2), int(self.shape[1] / 2), :]
        return p0, p1, p2

    def imshow(self, axis = 2):
        assert is_notebook()
        from matplotlib import pyplot as plt
        plt.imshow(self.central_slices[axis])


@dataclass
class Listmode:
    data: np.ndarray
    lors: LORs

    @property
    def vproj(self):
        return self.data

    @classmethod
    def from_scanner(cls, scanner):
        _lors = LORs.from_scanner(scanner)
        _data = np.ones((len(_lors),), dtype = np.float32)
        return cls(_data, _lors)

    def __len__(self):
        return len(self.lors)

    @classmethod
    def from_lors(cls, lors: LORs):
        return cls(np.ones((len(lors),), dtype = np.float32), lors)

    def to_sinogram(self, scanner: PETCylindricalScanner):
        iblock1 = _position_to_block_index(self.lors.data[:, :3], scanner)
        iblock2 = _position_to_block_index(self.lors.data[:, 3:], scanner)
        iring1 = _position_to_thin_ring_index(self.lors.data[:, :3], scanner)
        iring2 = _position_to_thin_ring_index(self.lors.data[:, 3:], scanner)
        _fst = _rotate_to_block0(self.lors.data[:, :3], scanner, iblock1)
        iy1 = _position_to_y_index_per_block(_fst, scanner)
        _snd = _rotate_to_block0(self.lors.data[:, 3:], scanner, iblock2)
        iy2 = _position_to_y_index_per_block(_snd, scanner)
        row = iy1 + scanner.blocks.shape[1] * iblock1 + scanner.nb_detectors_per_thin_ring * iring1
        col = iy2 + scanner.blocks.shape[1] * iblock2 + scanner.nb_detectors_per_thin_ring * iring2

        return Sinogram(sparse.csr_matrix((self.data, (row, col)), shape = (scanner.nb_detectors,
                                                                            scanner.nb_detectors
                                                                            )), scanner)

    def compress(self, scanner: PETCylindricalScanner):
        return self.to_sinogram(scanner).to_listmode()


def _position_to_block_index(pos, scanner: PETCylindricalScanner):
    xc, yc = pos[:, 0], pos[:, 1]
    return np.round(np.arctan2(yc, xc) / scanner.angle_per_block).astype(
        int) % scanner.nb_blocks_per_ring


def _position_to_thin_ring_index(pos, scanner: PETCylindricalScanner):
    zc = pos[:, 2]
    return np.floor((zc + scanner.axial_length / 2) / scanner.blocks.unit_size[2]).astype(int)


def _rotate_to_block0(pos, scanner: PETCylindricalScanner, iblock):
    angle = iblock * scanner.angle_per_block
    _pos = np.zeros(pos.shape)
    xc, yc = pos[:, 0], pos[:, 1]
    _pos[:, 0] = xc * np.cos(angle) + yc * np.sin(angle)
    _pos[:, 1] = -xc * np.sin(angle) + yc * np.cos(angle)
    _pos[:, 2] = pos[:, 2]
    return _pos


def _position_to_y_index_per_block(pos, scanner: PETCylindricalScanner):
    return np.round((pos[:, 1] + scanner.blocks.size[1] / 2) // scanner.blocks.unit_size[1]).astype(
        int)


@jit
def _mesh_detector(scanner):
    R = (scanner.inner_radius + scanner.outer_radius) / 2
    theta = 2 * np.pi / scanner.nb_blocks_per_ring * np.arange(scanner.nb_blocks_per_ring)

    x = np.ones(scanner.blocks.shape[1], ) * R
    y = (np.arange(scanner.blocks.shape[1]) - scanner.blocks.shape[1] / 2 + 0.5) * \
        scanner.blocks.unit_size[1]
    z = (np.arange(scanner.blocks.shape[2]) - scanner.blocks.shape[2] / 2 + 0.5) * \
        scanner.blocks.unit_size[2]
    xx = np.kron(x, [1] * scanner.nb_blocks_per_ring)
    yy = np.kron(y, [1] * scanner.nb_blocks_per_ring)
    theta1 = np.kron(theta, [[1]] * scanner.blocks.shape[1]).ravel()
    xx1 = xx * np.cos(theta1) - yy * np.sin(theta1)
    yy1 = xx * np.sin(theta1) + yy * np.cos(theta1)

    xd = np.kron(xx1, [[1]] * scanner.blocks.shape[2]).ravel()
    yd = np.kron(yy1, [[1]] * scanner.blocks.shape[2]).ravel()
    zd = np.kron(z, [1] * scanner.blocks.shape[1] * scanner.nb_blocks_per_ring).ravel()

    lors = np.zeros((scanner.nb_detectors * scanner.nb_detectors, 6), dtype = np.float32)
    for i in range(scanner.nb_detectors):
        for j in range(scanner.nb_detectors):
            j1 = j
            lors[i * scanner.nb_detectors + j, :] = [xd[i], yd[i], zd[i], xd[j1], yd[j1], zd[j1]]
    return lors


@dataclass
class UMap(Image):
    pass


@dataclass
class Emap_MLEM(Image):
    # TODO correct emap generations in ring mode.
    @classmethod
    def from_scanner(cls, scanner: PETCylindricalScanner, bprojector, mode = 'thin-ring'):
        emap = cls(np.zeros(bprojector.shape, np.float32), bprojector.center, bprojector.size)

        # if not np.abs(scanner.blocks.unit_size[2] - bprojector.unit_size[2]) < _tiny:
        if mode == 'full':
            _emap = bprojector(Listmode.from_lors(LORs.from_scanner(scanner)))
            return cls(_emap.data, _emap.center, _emap.size)
        elif mode == 'ring-full':
            _lors0 = LORs.from_scanner_ring(scanner, 0, 0)
            for i in _tqdm(np.arange(scanner.nb_rings)):
                for j in _tqdm(np.arange(scanner.nb_rings)):
                    if i > j:
                        continue
                    _lors = deepcopy(_lors0)
                    _lors.data[:, 2] += i * scanner.blocks.size[2] + i * scanner.gap - \
                                        scanner.axial_length / 2
                    _lors.data[:, 5] += j * scanner.blocks.size[2] + j * scanner.gap - \
                                        scanner.axial_length / 2
                    _emap = bprojector(Listmode.from_lors(_lors))
                    if i == j:
                        emap = emap + _emap
                    else:
                        emap = emap + _emap * 2
            return emap
        elif mode == 'ring':
            raise ValueError
            _lors0 = LORs.from_scanner_ring(scanner, 0, 0)
            for d in _tqdm(np.arange(0, scanner.nb_rings)):
                _lors = deepcopy(_lors0)
                _lors.data[:, 5] += d * scanner.blocks.size[2] + d * scanner.gap
                _emap = bprojector(Listmode.from_lors(_lors))
                for i in np.arange(scanner.nb_rings):
                    j = i + d
                    if not 0 <= j < scanner.nb_rings:
                        continue
                    if d == 0:
                        emap = emap + _emap.shift(
                            [0, 0, i * scanner.blocks.size[
                                2] + i * scanner.gap - scanner.axial_length / 2])
                    else:
                        emap = emap + _emap.shift(
                            [0, 0, i * scanner.blocks.size[
                                2] + i * scanner.gap - scanner.axial_length / 2]) * 2
            return emap
        elif mode == 'thin-ring-full':
            _lors0 = LORs.from_scanner_thin_ring(scanner, 0, 0)
            for i in _tqdm(np.arange(scanner.nb_thin_rings)):
                for j in np.arange(scanner.nb_thin_rings):
                    if i > j:
                        continue
                    _lors = deepcopy(_lors0)
                    _lors.data[:, 2] += i * scanner.blocks.unit_size[2] + i // \
                                        scanner.blocks.shape[
                                            2] * scanner.gap - scanner.axial_length / 2
                    _lors.data[:, 5] += j * scanner.blocks.unit_size[2] + j // \
                                        scanner.blocks.shape[
                                            2] * scanner.gap - scanner.axial_length / 2
                    _emap = bprojector(Listmode.from_lors(_lors))
                    if i == j:
                        emap = emap + _emap
                    else:
                        emap = emap + _emap * 2
            return emap
        elif mode == 'thin-ring':
            _lors0 = LORs.from_scanner_thin_ring(scanner, 0, 0)
            for d in _tqdm(np.arange(0, scanner.nb_thin_rings)):
                _lors = deepcopy(_lors0)
                _lors.data[:, 5] += d * scanner.blocks.unit_size[2]
                _emap = bprojector(Listmode.from_lors(_lors))
                for i in np.arange(scanner.nb_thin_rings):
                    j = i + d
                    if not 0 <= j < scanner.nb_thin_rings:
                        continue
                    if d == 0:
                        emap = emap + _emap.shift(
                            [0, 0, i * scanner.blocks.unit_size[2] - scanner.axial_length / 2])
                    else:
                        emap = emap + _emap.shift(
                            [0, 0, i * scanner.blocks.unit_size[2] - scanner.axial_length / 2]) * 2
            return emap
        else:
            raise NotImplementedError

    @classmethod
    def from_scanner_full(cls, scanner: PETCylindricalScanner, bprojector):
        _emap = bprojector(Listmode.from_lors(LORs.from_scanner(scanner)))
        return cls(_emap.data, _emap.center, _emap.size)


TYPE_BIND.update({'Block': Block, 'PETCylindricalScanner': PETCylindricalScanner,
                  'LORs': LORs, 'Sinogram': Sinogram, 'Image': Image, 'Listmode': Listmode,
                  'UMap': UMap, 'Emap_MLEM': Emap_MLEM})



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: __init__.py
@date: 2/26/2019
@desc:
'''
__all__ = (
    'ProjectorSiddon', 'ProjectorDistanceDriven', 'BackProjectorSiddon',
    'BackProjectorDistanceDriven')
from srfnef.templating import TYPE_BIND
from ._bproj_distance_driven import BackProjectorDistanceDriven
from ._bproj_siddon import BackProjectorSiddon
from ._proj_distance_driven import ProjectorDistanceDriven
from ._proj_siddon import ProjectorSiddon

TYPE_BIND.update({'BackProjectorDistanceDriven': BackProjectorDistanceDriven,
                  'BackProjectorSiddon': BackProjectorSiddon,
                  'ProjectorDistanceDriven': ProjectorDistanceDriven,
                  'ProjectorSiddon': ProjectorSiddon})



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: cli.py
@date: 3/4/2019
@desc:
'''
import click


#
# @click.command()
# @click.option('-r', '--recon', type = str, help = 'reconstructing scheme file, e.g. mlem.hdf5')
# @click.option('-l', '--listmode', type = str, help = 'listmode data file, e.g., listmode.hdf5')
# @click.option('-o', '--output', type = (str, int), help = 'output path, e.g. ./output?.hdf5, 5)',
#               default = ('./temp_recon_?.hdf5', 0))
# def recon(recon, listmode, output):
#     click.echo('Reconstructing...')
#     import srfnef as nef
#     mlem = nef.load(recon)
#     mlem1 = mlem.register(nef.templating.saver(output[0], output[1]))
#     listmode = nef.load(listmode)
#     recon_img = mlem1(listmode)
#     nef.save(output[0].replace('?', 'final'), recon_img)


#
# if __name__ == '__main__':
#     recon()


@click.command()
@click.option('-m', '--mode', type = click.Choice(['auto', 'generate']), default = 'auto')
@click.option('-f', '--function', type = click.Path(exists = True), help = 'filetion file path, '
                                                                           'e.g., ./mlem.hdf5')
@click.option('-i', '--input', type = click.Path(exists = True), help = 'input file path, e.g., '
                                                                        './listmode.hdf5')
@click.option('-o', '--output', type = click.Path(), help = 'output file path, e.g, '
                                                            './recon_image.hdf5', default =
              './result.hdf5')
def cli(mode, function, input, output):
    import srfnef as nef
    if mode == 'auto':
        func = nef.load(click.format_filename(function))
        click.echo('Running ' + func.split(('.', '/')[-2]) + ' process...')
        _in = nef.load(click.format_filename(input))
        _out = func(_in)
        nef.save(click.format_filename(output), _out)
    elif mode == 'generate':
        pass


if __name__ == '__main__':
    cli()



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: BackProjectors.py
@date: 1/7/2019
@desc:
'''

import math

import attr
import numpy as np
from numba import jit, cuda

from srfnef.data_types import Image, Listmode
from srfnef.func_types import BackProjector
from srfnef.templating import funcclass

running_env = jit(nopython = True, parallel = True)


@running_env
def _kernel_bproj_3d_siddon(i, fst, snd, vproj, unit_size, center, image):
    dx, dy, dz = unit_size
    dz /= dx
    x1, y1, z1 = (fst[i, 0] - center[0]) / dx, (fst[i, 1] - center[1]) / dx, (
            fst[i, 2] - center[2]) / dx
    x2, y2, z2 = (snd[i, 0] - center[0]) / dx, (snd[i, 1] - center[1]) / dx, (
            snd[i, 2] - center[2]) / dx
    dx = 1.
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if (xd ** 2 + yd ** 2) ** 0.5 < 10:
        return
    nx, ny, nz = image.shape[0], image.shape[1], image.shape[2]
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2
    L = (xd ** 2 + yd ** 2 + zd ** 2) ** 0.5
    if abs(xd) > abs(yd):
        ky = yd / xd
        kz = zd / xd
        for ix in range(nx):
            xx1 = ix - nx2
            xx2 = xx1 + 1
            if ky >= 0:
                yy1 = y1 + ky * (xx1 - x1) + ny2
                yy2 = y1 + ky * (xx2 - x1) + ny2
            else:
                yy1 = y1 + ky * (xx2 - x1) + ny2
                yy2 = y1 + ky * (xx1 - x1) + ny2
            cy1 = math.floor(yy1)
            cy2 = math.floor(yy2)

            if kz >= 0:
                zz1 = (z1 + kz * (xx1 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx2 - x1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (xx2 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx1 - x1)) / dz + nz2
            cz1 = math.floor(zz1)
            cz2 = math.floor(zz2)
            if cy1 == cy2:
                if 0 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            iy = cy1
                            iz = cz1
                            image[ix, iy, iz] += vproj[i] * (
                                    1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                iy = cy1
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * rz * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            if cz2 < nz:
                                iy = cy1
                                iz = cz2
                                image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
            else:
                if -1 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            if cy1 >= 0:
                                iy = cy1
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * ry * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            if cy2 < ny:
                                iy = cy2
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * (1 - ry) * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                    else:
                        if -1 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if ry > rz:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * rz * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cy1 >= 0 and cz2 < nz:
                                    iy = cy1
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (ry - rz) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (1 - ry) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            else:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * ry * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cy2 < ny and cz1 >= 0:
                                    iy = cy2
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * (rz - ry) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5 * dx / L / L
    else:
        kx = xd / yd
        kz = zd / yd
        for iy in range(ny):
            yy1 = iy - ny2
            yy2 = yy1 + 1
            if kx >= 0:
                xx1 = x1 + kx * (yy1 - y1) + nx2
                xx2 = x1 + kx * (yy2 - y1) + nx2
            else:
                xx1 = x1 + kx * (yy2 - y1) + nx2
                xx2 = x1 + kx * (yy1 - y1) + nx2
            cx1 = math.floor(xx1)
            cx2 = math.floor(xx2)

            if kz >= 0:
                zz1 = (z1 + kz * (yy1 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy2 - y1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (yy2 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy1 - y1)) / dz + nz2
            cz1 = math.floor(zz1)
            cz2 = math.floor(zz2)
            if cx1 == cx2:
                if 0 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ix = cx1
                            iz = cz1
                            image[ix, iy, iz] += vproj[i] * (
                                    1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                ix = cx1
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * rz * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            if cz2 < nz:
                                ix = cx1
                                iz = cz2
                                image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
            else:
                if -1 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            if cx1 >= 0:
                                ix = cx1
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * rx * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            if cx2 < nx:
                                ix = cx2
                                iz = cz1
                                image[ix, iy, iz] += vproj[i] * (1 - rx) * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                    else:
                        if -1 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if rx > rz:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * rz * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cx1 >= 0 and cz2 < nz:
                                    ix = cx1
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (rx - rz) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (1 - rx) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                            else:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * rx * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cx2 < nx and cz1 >= 0:
                                    ix = cx2
                                    iz = cz1
                                    image[ix, iy, iz] += vproj[i] * (rz - rx) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L
                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5 * dx / L / L


def bproj_siddon(vproj, lors, unit_size, center, shape):
    image = np.zeros(shape, dtype = np.float32)
    for i in range(lors.shape[0]):
        _kernel_bproj_3d_siddon(i, lors[:, :3], lors[:, 3:6], vproj, unit_size, center,
                                image)
    return image


@cuda.jit(device = True)
def _kernel_bproj_3d_siddon_cuda(i, fst, snd, vproj, unit_size, center, image):
    dx, dy, dz = unit_size
    dz /= dx
    x1, y1, z1 = (fst[i, 0] - center[0]) / dx, (fst[i, 1] - center[1]) / dx, (
            fst[i, 2] - center[2]) / dx
    x2, y2, z2 = (snd[i, 0] - center[0]) / dx, (snd[i, 1] - center[1]) / dx, (
            snd[i, 2] - center[2]) / dx
    dx = 1.
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if (xd ** 2 + yd ** 2) ** 0.5 < 10:
        return
    nx, ny, nz = image.shape[0], image.shape[1], image.shape[2]
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2
    # L = (xd ** 2 + yd ** 2 + zd ** 2) ** 0.5
    L = 1.0
    if abs(xd) > abs(yd):
        ky = yd / xd
        kz = zd / xd
        for ix in range(nx):
            xx1 = ix - nx2
            xx2 = xx1 + 1
            if ky >= 0:
                yy1 = y1 + ky * (xx1 - x1) + ny2
                yy2 = y1 + ky * (xx2 - x1) + ny2
            else:
                yy1 = y1 + ky * (xx2 - x1) + ny2
                yy2 = y1 + ky * (xx1 - x1) + ny2
            cy1 = math.floor(yy1)
            cy2 = math.floor(yy2)

            if kz >= 0:
                zz1 = (z1 + kz * (xx1 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx2 - x1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (xx2 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx1 - x1)) / dz + nz2
            cz1 = math.floor(zz1)
            cz2 = math.floor(zz2)
            if cy1 == cy2:
                if 0 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            iy = cy1
                            iz = cz1
                            cuda.atomic.add(image, (ix, iy, iz),
                                            vproj[i] * (1 + ky ** 2 + kz ** 2) ** 0.5)
                            # image[ix, iy, iz] += vproj[i] * (1 + ky ** 2 + kz ** 2) ** 0.5
                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                iy = cy1
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz),
                                                vproj[i] * rz * (
                                                        1 + ky ** 2 + kz ** 2) ** 0.5)
                                # image[ix, iy, iz] += vproj[i] * rz * (1 + ky ** 2 + kz ** 2) ** 0.5
                            if cz2 < nz:
                                iy = cy1
                                iz = cz2
                                cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rz) * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5)
                                # image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                #         1 + ky ** 2 + kz ** 2) ** 0.5
            else:
                if -1 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            if cy1 >= 0:
                                iy = cy1
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz),
                                                vproj[i] * ry * (
                                                        1 + ky ** 2 + kz ** 2) ** 0.5)
                                # image[ix, iy, iz] += vproj[i] * ry * (1 + ky ** 2 + kz ** 2) ** 0.5
                            if cy2 < ny:
                                iy = cy2
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - ry) * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5)
                                # image[ix, iy, iz] += vproj[i] * (1 - ry) * (
                                #         1 + ky ** 2 + kz ** 2) ** 0.5
                    else:
                        if -1 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if ry > rz:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * rz * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * rz * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
                                if cy1 >= 0 and cz2 < nz:
                                    iy = cy1
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (ry - rz) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * (ry - rz) * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - ry) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * (1 - ry) * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
                            else:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * ry * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * ry * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
                                if cy2 < ny and cz1 >= 0:
                                    iy = cy2
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (rz - ry) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * (rz - ry) * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rz) * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5)
                                    # image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                    #         1 + ky ** 2 + kz ** 2) ** 0.5
    else:
        kx = xd / yd
        kz = zd / yd
        for iy in range(ny):
            yy1 = iy - ny2
            yy2 = yy1 + 1
            if kx >= 0:
                xx1 = x1 + kx * (yy1 - y1) + nx2
                xx2 = x1 + kx * (yy2 - y1) + nx2
            else:
                xx1 = x1 + kx * (yy2 - y1) + nx2
                xx2 = x1 + kx * (yy1 - y1) + nx2
            cx1 = math.floor(xx1)
            cx2 = math.floor(xx2)

            if kz >= 0:
                zz1 = (z1 + kz * (yy1 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy2 - y1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (yy2 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy1 - y1)) / dz + nz2
            cz1 = math.floor(zz1)
            cz2 = math.floor(zz2)
            if cx1 == cx2:
                if 0 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ix = cx1
                            iz = cz1
                            cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 + kx ** 2 + kz **
                                                                             2) ** 0.5)
                            # image[ix, iy, iz] += vproj[i] * (1 + kx ** 2 + kz ** 2) ** 0.5
                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                ix = cx1
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz), vproj[i] * rz * (1 + kx ** 2
                                                                                      + kz ** 2) ** 0.5)
                                # image[ix, iy, iz] += vproj[i] * rz * (1 + kx ** 2 + kz ** 2) ** 0.5
                            if cz2 < nz:
                                ix = cx1
                                iz = cz2
                            cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rz) * (
                                    1 + kx ** 2 + kz ** 2) ** 0.5)
                            #         image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                            # 1 + kx ** 2 + kz ** 2) ** 0.5
            else:
                if -1 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            if cx1 >= 0:
                                ix = cx1
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz), vproj[i] * rx * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5)
                            #     image[ix, iy, iz] += vproj[i] * rx * (
                            # 1 + kx ** 2 + kz ** 2) ** 0.5
                            if cx2 < nx:
                                ix = cx2
                                iz = cz1
                                cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rx) * (
                                        1 + kx ** 2 + kz ** 2) ** 0.5)
                            #         image[ix, iy, iz] += vproj[i] * (1 - rx) * (
                            # 1 + kx ** 2 + kz ** 2) ** 0.5
                    else:
                        if -1 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if rx > rz:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * rz * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * rz * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5
                                if cx1 >= 0 and cz2 < nz:
                                    ix = cx1
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (rx - rz) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * (rx - rz) * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5
                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rx) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * (1 - rx) * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5
                            else:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * rx * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * rx * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5
                                if cx2 < nx and cz1 >= 0:
                                    ix = cx2
                                    iz = cz1
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (rz - rx) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * (rz - rx) * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5
                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    cuda.atomic.add(image, (ix, iy, iz), vproj[i] * (1 - rz) * (
                                            1 + kx ** 2 + kz ** 2) ** 0.5)
                                    #         image[ix, iy, iz] += vproj[i] * (1 - rz) * (
                                    # 1 + kx ** 2 + kz ** 2) ** 0.5


@cuda.jit
def bproj_siddon_cuda(vproj, lors, unit_size, center, image):
    i = cuda.grid(1)
    if i >= lors.shape[0]:
        return
    _kernel_bproj_3d_siddon_cuda(i, lors[:, :3], lors[:, 3:6], vproj, unit_size, center,
                                 image)


@funcclass
class BackProjectorSiddon(BackProjector):
    shape: np.ndarray = attr.ib(converter = np.array)
    center: np.ndarray = attr.ib(converter = np.array)
    size: np.ndarray = attr.ib(converter = np.array)
    device: str = 'gpu'

    def __call__(self, listmode: Listmode):
        if self.device == 'cpu':
            image_data = bproj_siddon(listmode.data, listmode.lors.data, self.size /
                                      self.shape, self.center, self.shape)
        elif self.device == 'gpu':
            blockdim = (256,)
            griddim = (1 + int(len(listmode.lors) / blockdim[0]),)
            image_data = np.zeros(self.shape, dtype = np.float32)
            bproj_siddon_cuda[griddim, blockdim](listmode.data, listmode.lors.data, self.size /
                                                 self.shape, self.center, image_data)
        else:
            raise NotImplementedError
        return Image(image_data, self.center, self.size)

    @property
    def unit_size(self):
        return self.size / self.shape



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: op.py
@date: 3/4/2019
@desc:
'''

import math

from numba import float32, guvectorize


@guvectorize([(float32[:, :, :], float32, float32[:, :, :])], '(m,n,p),()->(m,n,p)',
             target = 'cuda')
def weighted_add(x, wei, y):
    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                y[i, j, k] += x[i, j, k] * wei


@guvectorize([(float32[:, :, :], float32, float32[:, :, :])], '(m,n,p),()->(m,n,p)',
             target = 'cuda')
def shift_ind_z(x, off, y):
    '''
    move a 3d image x with off forward and added on a same-shaped image y
    '''

    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                ind1 = int(math.floor(k - off))
                ind2 = ind1 + 1
                wt2 = k - off - ind1
                wt1 = 1.0 - wt2
                print(k, ind1, ind2, wt1, wt2)
                if 0 <= ind1 and ind2 < x.shape[2]:
                    y[i, j, k] += x[i, j, ind1] * wt1 + x[i, j, ind2] * wt2
                elif ind2 == 0:
                    y[i, j, k] += x[i, j, 0] * wt2
                elif ind1 == x.shape[2] - 1:
                    y[i, j, k] += x[i, j, -1] * wt1
                else:
                    pass



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: _proj_siddon.py
@date: 1/21/2019
@desc:
'''
import numpy as np
from numba import jit, cuda

from srfnef.data_types import Image, Listmode, LORs
from srfnef.func_types import Projector
from srfnef.templating import funcclass

running_env = jit(nopython = True, parallel = True)


@running_env
def _kernel_proj_3d_siddon(i, image, fst, snd, unit_size, center, vproj):
    dx, dy, dz = unit_size
    dz /= dx
    x1, y1, z1 = (fst[i, 0] - center[0]) / dx, (fst[i, 1] - center[1]) / dx, (
            fst[i, 2] - center[2]) / dx
    x2, y2, z2 = (snd[i, 0] - center[0]) / dx, (snd[i, 1] - center[1]) / dx, (
            snd[i, 2] - center[2]) / dx
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if (xd ** 2 + yd ** 2) ** 0.5 < 10:
        return
    nx, ny, nz = image.shape[0], image.shape[1], image.shape[2]
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2
    L = (xd ** 2 + yd ** 2 + zd ** 2) ** 0.5
    if abs(xd) > abs(yd):
        ky = yd / xd
        kz = zd / xd
        for ix in range(nx):
            xx1 = ix - nx2
            xx2 = xx1 + 1
            if ky >= 0:
                yy1 = y1 + ky * (xx1 - x1) + ny2
                yy2 = y1 + ky * (xx2 - x1) + ny2
            else:
                yy1 = y1 + ky * (xx2 - x1) + ny2
                yy2 = y1 + ky * (xx1 - x1) + ny2
            cy1 = int(yy1)
            cy2 = int(yy2)

            if kz >= 0:
                zz1 = (z1 + kz * (xx1 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx2 - x1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (xx2 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx1 - x1)) / dz + nz2
            cz1 = int(zz1)
            cz2 = int(zz2)
            if cy1 == cy2:
                if 0 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            iy = cy1
                            iz = cz1
                            weight = (1 + ky ** 2 + kz ** 2) ** 0.5
                            vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                iy = cy1
                                iz = cz1
                                weight = rz * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cz2 < nz:
                                iy = cy1
                                iz = cz2
                                weight = (1 - rz) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

            else:
                if -1 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            if cy1 >= 0:
                                iy = cy1
                                iz = cz1
                                weight = ry * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cy2 < ny:
                                iy = cy2
                                iz = cz1
                                weight = (1 - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if ry > rz:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    weight = rz * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy1 >= 0 and cz2 < nz:
                                    iy = cy1
                                    iz = cz2
                                    weight = (ry - rz) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    weight = (1 - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                            else:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    weight = ry * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz1 >= 0:
                                    iy = cy2
                                    iz = cz1
                                    weight = (rz - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    weight = (1 - rz) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

    else:
        kx = xd / yd
        kz = zd / yd
        for iy in range(ny):
            yy1 = iy - ny2
            yy2 = yy1 + 1
            if kx >= 0:
                xx1 = x1 + kx * (yy1 - y1) + nx2
                xx2 = x1 + kx * (yy2 - y1) + nx2
            else:
                xx1 = x1 + kx * (yy2 - y1) + nx2
                xx2 = x1 + kx * (yy1 - y1) + nx2
            cx1 = int(xx1)
            cx2 = int(xx2)

            if kz >= 0:
                zz1 = (z1 + kz * (yy1 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy2 - y1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (yy2 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy1 - y1)) / dz + nz2
            cz1 = int(zz1)
            cz2 = int(zz2)
            if cx1 == cx2:
                if 0 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ix = cx1
                            iz = cz1
                            weight = (1 + kx ** 2 + kz ** 2) ** 0.5
                            vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                ix = cx1
                                iz = cz1
                                weight = rz * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cz2 < nz:
                                ix = cx1
                                iz = cz2
                                weight = (1 - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

            else:
                if -1 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            if cx1 >= 0:
                                ix = cx1
                                iz = cz1
                                weight = rx * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cx2 < nx:
                                ix = cx2
                                iz = cz1
                                weight = (1 - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if rx > rz:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    weight = rz * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx1 >= 0 and cz2 < nz:
                                    ix = cx1
                                    iz = cz2
                                    weight = (rx - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    weight = (1 - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                            else:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    weight = rx * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz1 >= 0:
                                    ix = cx2
                                    iz = cz1
                                    weight = (rz - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    weight = (1 - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

    # vproj[i] *= dx
    if L > 0:
        vproj[i] /= L * L


@running_env
def proj_siddon(image, lors, unit_size, center):
    vproj = np.zeros((lors.shape[0],), dtype = np.float32)

    for i in range(lors.shape[0]):
        _kernel_proj_3d_siddon(i, image, lors[:, :3], lors[:, 3:6], unit_size, center, vproj)
    return vproj


@cuda.jit(device = True)
def _kernel_proj_3d_siddon_cuda(i, image, fst, snd, unit_size, center, vproj):
    dx, dy, dz = unit_size
    dz /= dx
    x1, y1, z1 = (fst[0] - center[0]) / dx, (fst[1] - center[1]) / dx, \
                 (fst[2] - center[2]) / dx
    x2, y2, z2 = (snd[0] - center[0]) / dx, (snd[1] - center[1]) / dx, \
                 (snd[2] - center[2]) / dx
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if (xd ** 2 + yd ** 2) ** 0.5 < 10:
        return
    nx, ny, nz = image.shape[0], image.shape[1], image.shape[2]
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2
    L = (xd ** 2 + yd ** 2 + zd ** 2) ** 0.5
    if abs(xd) > abs(yd):
        ky = yd / xd
        kz = zd / xd
        for ix in range(nx):
            xx1 = ix - nx2
            xx2 = xx1 + 1.0
            if ky >= 0:
                yy1 = y1 + ky * (xx1 - x1) + ny2
                yy2 = y1 + ky * (xx2 - x1) + ny2
            else:
                yy1 = y1 + ky * (xx2 - x1) + ny2
                yy2 = y1 + ky * (xx1 - x1) + ny2
            cy1 = int(yy1) if yy1 >= 0.0 else int(yy1) - 1
            cy2 = int(yy2) if yy2 >= 0.0 else int(yy2) - 1

            if kz >= 0:
                zz1 = (z1 + kz * (xx1 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx2 - x1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (xx2 - x1)) / dz + nz2
                zz2 = (z1 + kz * (xx1 - x1)) / dz + nz2
            cz1 = int(zz1) if zz1 >= 0.0 else int(zz1) - 1
            cz2 = int(zz2) if zz2 >= 0.0 else int(zz2) - 1

            if cy1 == cy2:
                if 0 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            iy = cy1
                            iz = cz1
                            weight = (1 + ky ** 2 + kz ** 2) ** 0.5
                            vproj[i] += image[ix, iy, iz] * weight
                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                iy = cy1
                                iz = cz1
                                weight = rz * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cz2 < nz:
                                iy = cy1
                                iz = cz2
                                weight = (1 - rz) * (
                                        1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight
            else:
                if -1 <= cy1 < ny:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            if cy1 >= 0:
                                iy = cy1
                                iz = cz1
                                weight = ry * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cy2 < ny:
                                iy = cy2
                                iz = cz1
                                weight = (1 - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                    else:

                        if -1 <= cz1 < nz:
                            ry = (cy2 - yy1) / (yy2 - yy1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if ry > rz:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    weight = rz * (
                                            1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy1 >= 0 and cz2 < nz:
                                    iy = cy1
                                    iz = cz2
                                    weight = (ry - rz) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    weight = (1 - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                            else:
                                if cy1 >= 0 and cz1 >= 0:
                                    iy = cy1
                                    iz = cz1
                                    weight = ry * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz1 >= 0:
                                    iy = cy2
                                    iz = cz1
                                    weight = (rz - ry) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cy2 < ny and cz2 < nz:
                                    iy = cy2
                                    iz = cz2
                                    weight = (1 - rz) * (1 + ky ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

    else:

        kx = xd / yd
        kz = zd / yd
        for iy in range(ny):
            yy1 = iy - ny2
            yy2 = yy1 + 1
            if kx >= 0:
                xx1 = x1 + kx * (yy1 - y1) + nx2
                xx2 = x1 + kx * (yy2 - y1) + nx2
            else:
                xx1 = x1 + kx * (yy2 - y1) + nx2
                xx2 = x1 + kx * (yy1 - y1) + nx2
            cx1 = int(xx1) if xx1 >= 0.0 else int(xx1) - 1
            cx2 = int(xx2) if xx2 >= 0.0 else int(xx2) - 1

            if kz >= 0:
                zz1 = (z1 + kz * (yy1 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy2 - y1)) / dz + nz2
            else:
                zz1 = (z1 + kz * (yy2 - y1)) / dz + nz2
                zz2 = (z1 + kz * (yy1 - y1)) / dz + nz2
            cz1 = int(zz1) if zz1 >= 0.0 else int(zz1) - 1
            cz2 = int(zz2) if zz2 >= 0.0 else int(zz2) - 1
            if cx1 == cx2:
                if 0 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            ix = cx1
                            iz = cz1
                            weight = (1 + kx ** 2 + kz ** 2) ** 0.5
                            vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if cz1 >= 0:
                                ix = cx1
                                iz = cz1
                                weight = rz * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cz2 < nz:
                                ix = cx1
                                iz = cz2
                                weight = (1 - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

            else:
                if -1 <= cx1 < nx:
                    if cz1 == cz2:
                        if 0 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            if cx1 >= 0:
                                ix = cx1
                                iz = cz1
                                weight = rx * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                            if cx2 < nx:
                                ix = cx2
                                iz = cz1
                                weight = (1 - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                vproj[i] += image[ix, iy, iz] * weight

                    else:
                        if -1 <= cz1 < nz:
                            rx = (cx2 - xx1) / (xx2 - xx1)
                            rz = (cz2 - zz1) / (zz2 - zz1)
                            if rx > rz:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    weight = rz * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx1 >= 0 and cz2 < nz:
                                    ix = cx1
                                    iz = cz2
                                    weight = (rx - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    weight = (1 - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                            else:
                                if cx1 >= 0 and cz1 >= 0:
                                    ix = cx1
                                    iz = cz1
                                    weight = rx * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz1 >= 0:
                                    ix = cx2
                                    iz = cz1
                                    weight = (rz - rx) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

                                if cx2 < nx and cz2 < nz:
                                    ix = cx2
                                    iz = cz2
                                    weight = (1 - rz) * (1 + kx ** 2 + kz ** 2) ** 0.5
                                    vproj[i] += image[ix, iy, iz] * weight

    # vproj[i] *= dx
    if L > 0:
        vproj[i] /= L * L


@cuda.jit
def proj_siddon_cuda(image, lors, unit_size, center, vproj):
    # i, j = cuda.grid(2)
    # id = i + j * cuda.gridDim.x * cuda.blockDim.x
    # if id >= lors.shape[0]:
    #     return
    # _kernel_proj_3d_siddon_cuda(id, image, lors[id, :3], lors[id, 3:6], unit_size, center,
    #                             vproj)
    i = cuda.grid(1)
    if i >= lors.shape[0]:
        return
    _kernel_proj_3d_siddon_cuda(i, image, lors[i, :3], lors[i, 3:6], unit_size, center,
                                vproj)


@funcclass
class ProjectorSiddon(Projector):
    device: str = 'gpu'

    def __call__(self, image: Image, lors: LORs):
        if self.device == 'cpu':
            vproj = proj_siddon(image.data, lors.data, image.size / image.shape, image.center)
        elif self.device == 'gpu':
            blockdim = (256,)
            griddim = (1 + int(len(lors) / blockdim[0]),)
            vproj = np.zeros((len(lors),), dtype = np.float32)
            proj_siddon_cuda[griddim, blockdim](image.data, lors.data,
                                                image.size / image.shape, image.center,
                                                vproj)
            cuda.synchronize()
        else:
            raise NotImplementedError

        return Listmode(vproj, lors)



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: _bproj_distance_driven.py
@date: 1/21/2019
@desc:
'''

import math

import attr
import numpy as np
from numba import jit, cuda, float32

from srfnef.data_types import Image, Listmode, PETCylindricalScanner
from srfnef.func_types import BackProjector
from srfnef.templating import funcclass
from srfnef.utils import _pi

running_env = jit(nopython = True, parallel = True)


@jit
def position_to_block_angle(pos, nb_blocks):
    xc, yc, _ = pos
    angle_per_block = 2 * _pi / nb_blocks
    iangle = (angle_per_block / 2 + math.atan2(yc, xc)) // angle_per_block
    return iangle


@jit
def _kernel_bproj_3d_distance_driven(vproj, fst, snd, blockdy2, blockdz2, nb_blocks,
                                     center, unit_size, image):
    dx, dy, dz = unit_size
    cx, cy, cz = center
    x1, y1, z1 = fst
    x2, y2, z2 = snd
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if xd ** 2 + yd ** 2 < 100:
        return
    nx, ny, nz = image.shape
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2

    itheta1 = position_to_block_angle(fst, nb_blocks)
    itheta2 = position_to_block_angle(snd, nb_blocks)
    angle_per_block = 2 * _pi / nb_blocks
    # if abs(itheta1 % nb_blocks - itheta2 % nb_blocks) <= 1.:
    #     return
    ctheta1, stheta1 = math.cos(itheta1 * angle_per_block), math.sin(itheta1 * angle_per_block)
    ctheta2, stheta2 = math.cos(itheta2 * angle_per_block), math.sin(itheta2 * angle_per_block)
    x1t, x1b = x1 - stheta1 * blockdy2, x1 + stheta1 * blockdy2
    x2t, x2b = x2 + stheta2 * blockdy2, x2 - stheta2 * blockdy2
    y1t, y1b = y1 + ctheta1 * blockdy2, y1 - ctheta1 * blockdy2
    y2t, y2b = y2 - ctheta2 * blockdy2, y2 + ctheta2 * blockdy2

    z1t, z1b = z1 + blockdz2, z1 - blockdz2
    z2t, z2b = z2 + blockdz2, z2 - blockdz2
    if abs(xd) >= abs(yd):
        calpha = abs(xd) / (xd ** 2 + yd ** 2) ** 0.5
        cgamma = abs(xd) / (xd ** 2 + zd ** 2) ** 0.5
        for ix in range(nx):
            xc = (ix - nx2 + 0.5) * dx + cx
            yt = ((xc - x1t) * (y2t - y1t) / (x2t - x1t) + y1t - cy) / dy + ny2
            yb = ((xc - x1b) * (y2b - y1b) / (x2b - x1b) + y1b - cy) / dy + ny2
            if yb > yt:
                yt, yb = yb, yt
            if yb >= ny or yt < 0:
                continue
            zt = ((xc - x1t) * (z2t - z1t) / (x2t - x1t) + z1t - cz) / dz + nz2
            zb = ((xc - x1b) * (z2b - z1b) / (x2b - x1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for iy in range(max(0, int(yb)), min(ny, 1 + int(yt))):
                wy = min(iy + 1, yt) - max(iy, yb)
                # wy /= (yt - yb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    # vproj[i] += image[ix, iy, iz] * wy * wz / calpha / cgamma
                    # cuda.atomic.add(image, (ix, iy, iz), wy * wz / calpha / cgamma * vproj[i])
                    image[ix, iy, iz] += wy * wz * vproj
    else:
        calpha = abs(yd) / (yd ** 2 + xd ** 2) ** 0.5
        cgamma = abs(yd) / (yd ** 2 + zd ** 2) ** 0.5
        for iy in range(ny):
            yc = (iy - ny2 + 0.5) * dy + cy
            xt = ((yc - y1t) * (x2t - x1t) / (y2t - y1t) + x1t - cx) / dx + nx2
            xb = ((yc - y1b) * (x2b - x1b) / (y2b - y1b) + x1b - cx) / dx + nx2
            if xb > xt:
                xt, xb = xb, xt
            if xb >= nx or xt < 0:
                continue
            zt = ((yc - y1t) * (z2t - z1t) / (y2t - y1t) + z1t - cz) / dz + nz2
            zb = ((yc - y1b) * (z2b - z1b) / (y2b - y1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for ix in range(max(0, int(xb)), min(nx, 1 + int(xt))):
                wx = min(ix + 1, xt) - max(ix, xb)
                # wx /= (xt - xb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    # vproj[i] += image[ix, iy, iz] * wx * wz / calpha / cgamma
                    # cuda.atomic.add(image, (ix, iy, iz), wx * wz / calpha / cgamma * vproj[i])
                    image[ix, iy, iz] += wx * wz * vproj


# @jit
def bproj_distance_driven(vproj, lors, shape, blockdy2, blockdz2, nb_blocks, center, unit_size):
    image = np.zeros(shape, dtype = np.float32)
    for i in range(lors.shape[0]):
        _kernel_bproj_3d_distance_driven(vproj[i], lors[i, :3], lors[i, 3:],
                                         blockdy2, blockdz2, nb_blocks, center, unit_size, image)
    return image


@cuda.jit(device = True)
def position_to_block_angle_cuda(pos, nb_blocks):
    xc, yc, _ = pos
    angle_per_block = 2 * _pi / nb_blocks
    #     angle = round(math.atan2(yc, xc) // angle_per_block) * angle_per_block
    iangle = (angle_per_block / 2 + math.atan2(yc, xc)) // angle_per_block
    #     iangle = math.floor(angle)
    return iangle


@cuda.jit(device = True)
def _kernel_bproj_3d_distance_driven_cuda(i, vproj, fst, snd, blockdy2, blockdz2, nb_blocks,
                                          center, unit_size, image):
    dx, dy, dz = unit_size
    cx, cy, cz = center
    x1, y1, z1 = fst
    x2, y2, z2 = snd
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if xd ** 2 + yd ** 2 < 100:
        return
    nx, ny, nz = image.shape
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2

    itheta1 = position_to_block_angle_cuda(fst, nb_blocks)
    itheta2 = position_to_block_angle_cuda(snd, nb_blocks)
    angle_per_block = 2 * _pi / nb_blocks
    # if abs(itheta1 % nb_blocks - itheta2 % nb_blocks) <= 1.:
    #     return
    ctheta1, stheta1 = math.cos(itheta1 * angle_per_block), math.sin(itheta1 * angle_per_block)
    ctheta2, stheta2 = math.cos(itheta2 * angle_per_block), math.sin(itheta2 * angle_per_block)
    x1t, x1b = x1 - stheta1 * blockdy2, x1 + stheta1 * blockdy2
    x2t, x2b = x2 + stheta2 * blockdy2, x2 - stheta2 * blockdy2
    y1t, y1b = y1 + ctheta1 * blockdy2, y1 - ctheta1 * blockdy2
    y2t, y2b = y2 - ctheta2 * blockdy2, y2 + ctheta2 * blockdy2

    z1t, z1b = z1 + blockdz2, z1 - blockdz2
    z2t, z2b = z2 + blockdz2, z2 - blockdz2
    if abs(xd) >= abs(yd):
        calpha = abs(xd) / (xd ** 2 + yd ** 2) ** 0.5
        cgamma = abs(xd) / (xd ** 2 + zd ** 2) ** 0.5
        for ix in range(nx):
            xc = (ix - nx2 + 0.5) * dx + cx
            yt = ((xc - x1t) * (y2t - y1t) / (x2t - x1t) + y1t - cy) / dy + ny2
            yb = ((xc - x1b) * (y2b - y1b) / (x2b - x1b) + y1b - cy) / dy + ny2
            if yb > yt:
                yt, yb = yb, yt
            if yb >= ny or yt < 0:
                continue
            zt = ((xc - x1t) * (z2t - z1t) / (x2t - x1t) + z1t - cz) / dz + nz2
            zb = ((xc - x1b) * (z2b - z1b) / (x2b - x1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for iy in range(max(0, int(yb)), min(ny, 1 + int(yt))):
                wy = min(iy + 1, yt) - max(iy, yb)
                # wy /= (yt - yb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    # vproj[i] += image[ix, iy, iz] * wy * wz / calpha / cgamma
                    # cuda.atomic.add(image, (ix, iy, iz), wy * wz / calpha / cgamma * vproj[i])
                    cuda.atomic.add(image, (ix, iy, iz), wy * wz * vproj[i])
    else:
        calpha = abs(yd) / (yd ** 2 + xd ** 2) ** 0.5
        cgamma = abs(yd) / (yd ** 2 + zd ** 2) ** 0.5
        for iy in range(ny):
            yc = (iy - ny2 + 0.5) * dy + cy
            xt = ((yc - y1t) * (x2t - x1t) / (y2t - y1t) + x1t - cx) / dx + nx2
            xb = ((yc - y1b) * (x2b - x1b) / (y2b - y1b) + x1b - cx) / dx + nx2
            if xb > xt:
                xt, xb = xb, xt
            if xb >= nx or xt < 0:
                continue
            zt = ((yc - y1t) * (z2t - z1t) / (y2t - y1t) + z1t - cz) / dz + nz2
            zb = ((yc - y1b) * (z2b - z1b) / (y2b - y1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for ix in range(max(0, int(xb)), min(nx, 1 + int(xt))):
                wx = min(ix + 1, xt) - max(ix, xb)
                # wx /= (xt - xb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    # vproj[i] += image[ix, iy, iz] * wx * wz / calpha / cgamma
                    # cuda.atomic.add(image, (ix, iy, iz), wx * wz / calpha / cgamma * vproj[i])
                    cuda.atomic.add(image, (ix, iy, iz), wx * wz * vproj[i])


@cuda.jit
def bproj_distance_driven_cuda(vproj, lors, blockdy2, blockdz2, nb_blocks, center, unit_size,
                               image):
    i = cuda.grid(1)
    if i >= lors.shape[0]:
        return
    _kernel_bproj_3d_distance_driven_cuda(i, vproj, lors[i, :3], lors[i, 3:],
                                          blockdy2, blockdz2, nb_blocks, center, unit_size, image)


@funcclass
class BackProjectorDistanceDriven(BackProjector):
    shape: np.ndarray = attr.ib(converter = np.array)
    center: np.ndarray = attr.ib(converter = np.array)
    size: np.ndarray = attr.ib(converter = np.array)
    scanner: PETCylindricalScanner
    device: str = 'gpu'

    def __call__(self, listmode: Listmode):
        nb_blocks = int(self.scanner.nb_blocks_per_ring)
        blockdy2 = self.scanner.blocks.unit_size[1] / 2
        blockdz2 = self.scanner.blocks.unit_size[2] / 2
        _center = self.center
        _unit_size = self.size / self.shape

        if self.device == 'cpu':
            image_data = bproj_distance_driven(listmode.data, listmode.lors.data, self.shape,
                                               blockdy2, blockdz2, nb_blocks, _center, _unit_size)
        elif self.device == 'gpu':
            blockdim = (256,)
            griddim = (1 + int(len(listmode.lors) / blockdim[0]),)
            image_data = np.zeros(self.shape, dtype = np.float32)
            if not isinstance(listmode.data, np.ndarray):
                image_data = cuda.device_array(self.shape, dtype = float32)
            bproj_distance_driven_cuda[griddim, blockdim](listmode.data, listmode.lors.data,
                                                          blockdy2, blockdz2, nb_blocks, _center,
                                                          _unit_size, image_data)
        else:
            raise NotImplementedError
        return Image(image_data, self.center, self.size)

    @property
    def unit_size(self):
        return self.size / self.shape



from abc import abstractmethod

import attr
import numpy as np
from numba import jit

from .data_types import Image, Listmode, Emap_MLEM
from .templating import funcclass, noop, FuncClass
from .utils import _tqdm

running_env = jit(nopython = True, parallel = True)


@funcclass
class Projector:
    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


def projector_picker(method: str = 'siddon'):
    from .model_impl import ProjectorDistanceDriven, ProjectorSiddon
    if method == 'siddon':
        return ProjectorSiddon
    elif method == 'distance-driven':
        return ProjectorDistanceDriven
    else:
        raise NotImplementedError


@funcclass
class BackProjector:
    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


def bprojector_picker(method: str = 'siddon'):
    from .model_impl import BackProjectorSiddon, BackProjectorDistanceDriven
    if method == 'siddon':
        return BackProjectorSiddon
    elif method == 'distance-driven':
        return BackProjectorDistanceDriven
    else:
        raise NotImplementedError


@funcclass
class MLEM:
    n_iter: int
    projector: Projector
    bprojector: BackProjector
    emap: Emap_MLEM
    saver: FuncClass = attr.ib(default = noop)

    def __call__(self, listmode: Listmode):
        x = Image(np.ones(self.emap.shape, dtype = np.float32), self.emap.center,
                  self.emap.size)
        for ind in _tqdm(range(self.n_iter)):
            proj = self.projector(x, listmode.lors)
            bp = self.bprojector(listmode / proj)
            x = x * bp / self.emap
            self.saver(ind + 1, x)
        return x


from srfnef.templating import TYPE_BIND

TYPE_BIND.update({'Projector': Projector, 'BackProjector': BackProjector, 'MLEM': MLEM})



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: __init__.py.py
@date: 12/25/2018
@desc:
'''

from srfnef.app import cli
# from .correction.attenuation import UMapProjector
from . import model_impl, adapter, op
from .data_types import Block, PETCylindricalScanner, LORs, Sinogram, Image, Listmode, UMap, \
    Emap_MLEM
from .func_types import projector_picker, bprojector_picker, MLEM
from .templating import save, load
from .utils import _tqdm
# from .app import *



# encoding: utf-8
'''
@author: Minghao Guo
@contact: mh.guo0111@gmail.com
@software: srf_ct
@file: _proj_distance_driven.py
@date: 1/21/2019
@desc:
'''
import math

import numpy as np
from numba import jit, cuda, float32

from srfnef.data_types import Image, Listmode, LORs, PETCylindricalScanner
from srfnef.func_types import Projector
from srfnef.templating import funcclass
from srfnef.utils import _pi

running_env = jit(nopython = True, parallel = True)


@running_env
def position_to_block_angle(pos, nb_blocks):
    xc, yc, _ = pos
    angle_per_block = 2 * _pi / nb_blocks
    iangle = (angle_per_block / 2 + math.atan2(yc, xc)) / angle_per_block
    return iangle % nb_blocks


@running_env
def _kernel_proj_3d_distance_driven(i, image, fst, snd, blockdy2, blockdz2, unit_size, center,
                                    nb_blocks, vproj):
    dx, dy, dz = unit_size
    cx, cy, cz = center
    x1, y1, z1 = fst
    x2, y2, z2 = snd
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if xd ** 2 + yd ** 2 < 100:
        return
    nx, ny, nz = image.shape
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2

    itheta1 = position_to_block_angle(fst, nb_blocks)
    itheta2 = position_to_block_angle(snd, nb_blocks)
    angle_per_block = 2 * _pi / nb_blocks
    ctheta1, stheta1 = math.cos(itheta1 * angle_per_block), math.sin(itheta1 * angle_per_block)
    ctheta2, stheta2 = math.cos(itheta2 * angle_per_block), math.sin(itheta2 * angle_per_block)
    x1t, x1b = x1 - stheta1 * blockdy2, x1 + stheta1 * blockdy2
    x2t, x2b = x2 + stheta2 * blockdy2, x2 - stheta2 * blockdy2
    y1t, y1b = y1 + ctheta1 * blockdy2, y1 - ctheta1 * blockdy2
    y2t, y2b = y2 - ctheta2 * blockdy2, y2 + ctheta2 * blockdy2

    z1t, z1b = z1 + blockdz2, z1 - blockdz2
    z2t, z2b = z2 + blockdz2, z2 - blockdz2
    L1 = 0.
    L2 = 0.
    if abs(xd) >= abs(yd):
        calpha = abs(xd) / (xd ** 2 + yd ** 2) ** 0.5
        cgamma = abs(xd) / (xd ** 2 + zd ** 2) ** 0.5
        yt = (x1 - x1t) * (y2t - y1t) / (x2t - x1t) + y1t + ny2
        yb = (x1 - x1b) * (y2b - y1b) / (x2b - x1b) + y1b + ny2
        L1 += abs(yt - yb)
        yt = (x2 - x1t) * (y2t - y1t) / (x2t - x1t) + y1t + ny2
        yb = (x2 - x1b) * (y2b - y1b) / (x2b - x1b) + y1b + ny2
        L1 += abs(yt - yb)
        L1 *= abs(x2 - x1) / 2
        zt = ((x1 - x1t) * (z2t - z1t) / (x2t - x1t) + z1t) / dz + nz2
        zb = ((x1 - x1b) * (z2b - z1b) / (x2b - x1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        zt = ((x2 - x1t) * (z2t - z1t) / (x2t - x1t) + z1t) / dz + nz2
        zb = ((x2 - x1b) * (z2b - z1b) / (x2b - x1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        L2 *= abs(x2 - x1) / 2
        for ix in range(nx):
            xc = (ix - nx2 + 0.5) * dx + cx
            yt = ((xc - x1t) * (y2t - y1t) / (x2t - x1t) + y1t - cy) / dy + ny2
            yb = ((xc - x1b) * (y2b - y1b) / (x2b - x1b) + y1b - cy) / dy + ny2
            if yb > yt:
                yt, yb = yb, yt
            if yb >= ny or yt < 0:
                continue
            zt = ((xc - x1t) * (z2t - z1t) / (x2t - x1t) + z1t - cz) / dz + nz2
            zb = ((xc - x1b) * (z2b - z1b) / (x2b - x1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for iy in range(max(0, int(yb)), min(ny, 1 + int(yt))):
                wy = min(iy + 1, yt) - max(iy, yb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    vproj[i] += image[ix, iy, iz] * wy * wz
    else:
        calpha = abs(yd) / (yd ** 2 + xd ** 2) ** 0.5
        cgamma = abs(yd) / (yd ** 2 + zd ** 2) ** 0.5
        xt = (y1 - y1t) * (x2t - x1t) / (y2t - y1t) + x1t + nx2
        xb = (y1 - y1b) * (x2b - x1b) / (y2b - y1b) + x1b + nx2
        L1 += abs(xt - xb)
        xt = (y2 - y1t) * (x2t - x1t) / (y2t - y1t) + x1t + nx2
        xb = (y2 - y1b) * (x2b - x1b) / (y2b - y1b) + x1b + nx2
        L1 += abs(xt - xb)
        L1 *= abs(y2 - y1) / 2
        zt = ((y1 - y1t) * (z2t - z1t) / (y2t - y1t) + z1t) / dz + nz2
        zb = ((y1 - y1b) * (z2b - z1b) / (y2b - y1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        zt = ((y2 - y1t) * (z2t - z1t) / (y2t - y1t) + z1t) / dz + nz2
        zb = ((y2 - y1b) * (z2b - z1b) / (y2b - y1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        L2 *= abs(y2 - y1) / 2
        for iy in range(ny):
            yc = (iy - ny2 + 0.5) * dy + cy
            xt = ((yc - y1t) * (x2t - x1t) / (y2t - y1t) + x1t - cx) / dx + nx2
            xb = ((yc - y1b) * (x2b - x1b) / (y2b - y1b) + x1b - cx) / dx + nx2
            if xb > xt:
                xt, xb = xb, xt
            if xb >= nx or xt < 0:
                continue
            zt = ((yc - y1t) * (z2t - z1t) / (y2t - y1t) + z1t - cz) / dz + nz2
            zb = ((yc - y1b) * (z2b - z1b) / (y2b - y1b) + z1b - cz) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue

            for ix in range(max(0, int(xb)), min(nx, 1 + int(xt))):
                wx = min(ix + 1, xt) - max(ix, xb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    vproj[i] += image[ix, iy, iz] * wx * wz
    if L1 * L2 > 0:
        vproj[i] /= L1 * L2


@running_env
def proj_distance_driven(image, lors, blockdy2, blockdz2, unit_size, center, nb_blocks):
    vproj = np.zeros((lors.shape[0],), dtype = np.float32)
    for i in range(lors.shape[0]):
        _kernel_proj_3d_distance_driven(i, image, lors[i, :3], lors[i, 3:],
                                        blockdy2, blockdz2, unit_size, center, nb_blocks, vproj)
    return vproj


@cuda.jit(device = True)
def position_to_block_angle_cuda(pos, nb_blocks):
    xc, yc, _ = pos
    angle_per_block = 2 * _pi / nb_blocks
    #     angle = round(math.atan2(yc, xc) // angle_per_block) * angle_per_block
    iangle = (angle_per_block / 2 + math.atan2(yc, xc)) / angle_per_block
    #     iangle = math.floor(angle)
    return iangle % nb_blocks


@cuda.jit(device = True)
def _kernel_proj_3d_distance_driven_cuda(i, image, fst, snd, blockdy2, blockdz2, unit_size, center,
                                         nb_blocks, vproj):
    dx, dy, dz = unit_size
    dz /= dx
    x1, y1, z1 = (fst[0] - center[0]) / dx, (fst[1] - center[1]) / dx, \
                 (fst[2] - center[2]) / dx
    x2, y2, z2 = (snd[0] - center[0]) / dx, (snd[1] - center[1]) / dx, \
                 (snd[2] - center[2]) / dx
    xd, yd, zd = x2 - x1, y2 - y1, z2 - z1
    if xd ** 2 + yd ** 2 < 100:
        return
    nx, ny, nz = image.shape
    nx2, ny2, nz2 = nx / 2, ny / 2, nz / 2

    itheta1 = position_to_block_angle_cuda(fst, nb_blocks)
    itheta2 = position_to_block_angle_cuda(snd, nb_blocks)
    angle_per_block = 2 * _pi / nb_blocks

    ctheta1, stheta1 = math.cos(itheta1 * angle_per_block), math.sin(itheta1 * angle_per_block)
    ctheta2, stheta2 = math.cos(itheta2 * angle_per_block), math.sin(itheta2 * angle_per_block)
    x1t, x1b = x1 - stheta1 * blockdy2, x1 + stheta1 * blockdy2
    x2t, x2b = x2 + stheta2 * blockdy2, x2 - stheta2 * blockdy2
    y1t, y1b = y1 + ctheta1 * blockdy2, y1 - ctheta1 * blockdy2
    y2t, y2b = y2 - ctheta2 * blockdy2, y2 + ctheta2 * blockdy2

    z1t, z1b = z1 + blockdz2, z1 - blockdz2
    z2t, z2b = z2 + blockdz2, z2 - blockdz2
    L1 = 0.
    L2 = 0.
    if abs(xd) >= abs(yd):
        calpha = abs(xd) / (xd ** 2 + yd ** 2) ** 0.5
        cgamma = abs(xd) / (xd ** 2 + zd ** 2) ** 0.5
        yt = (x1 - x1t) * (y2t - y1t) / (x2t - x1t) + y1t + ny2
        yb = (x1 - x1b) * (y2b - y1b) / (x2b - x1b) + y1b + ny2
        L1 += abs(yt - yb)
        yt = (x2 - x1t) * (y2t - y1t) / (x2t - x1t) + y1t + ny2
        yb = (x2 - x1b) * (y2b - y1b) / (x2b - x1b) + y1b + ny2
        L1 += abs(yt - yb)
        L1 *= abs(x2 - x1) / 2
        zt = ((x1 - x1t) * (z2t - z1t) / (x2t - x1t) + z1t) / dz + nz2
        zb = ((x1 - x1b) * (z2b - z1b) / (x2b - x1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        zt = ((x2 - x1t) * (z2t - z1t) / (x2t - x1t) + z1t) / dz + nz2
        zb = ((x2 - x1b) * (z2b - z1b) / (x2b - x1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        L2 *= abs(x2 - x1) / 2

        for ix in range(nx):
            xc = ix - nx2 + 0.5
            yt = (xc - x1t) * (y2t - y1t) / (x2t - x1t) + y1t + ny2
            yb = (xc - x1b) * (y2b - y1b) / (x2b - x1b) + y1b + ny2
            if yb > yt:
                yt, yb = yb, yt
            if yb >= ny or yt < 0:
                continue
            zt = ((xc - x1t) * (z2t - z1t) / (x2t - x1t) + z1t) / dz + nz2
            zb = ((xc - x1b) * (z2b - z1b) / (x2b - x1b) + z1b) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for iy in range(max(0, int(yb)), min(ny, 1 + int(yt))):
                wy = min(iy + 1, yt) - max(iy, yb)
                # wy /= (yt - yb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    vproj[i] += image[ix, iy, iz] * wy * wz
    else:
        calpha = abs(yd) / (yd ** 2 + xd ** 2) ** 0.5
        cgamma = abs(yd) / (yd ** 2 + zd ** 2) ** 0.5
        xt = (y1 - y1t) * (x2t - x1t) / (y2t - y1t) + x1t + nx2
        xb = (y1 - y1b) * (x2b - x1b) / (y2b - y1b) + x1b + nx2
        L1 += abs(xt - xb)
        xt = (y2 - y1t) * (x2t - x1t) / (y2t - y1t) + x1t + nx2
        xb = (y2 - y1b) * (x2b - x1b) / (y2b - y1b) + x1b + nx2
        L1 += abs(xt - xb)
        L1 *= abs(y2 - y1) / 2
        zt = ((y1 - y1t) * (z2t - z1t) / (y2t - y1t) + z1t) / dz + nz2
        zb = ((y1 - y1b) * (z2b - z1b) / (y2b - y1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        zt = ((y2 - y1t) * (z2t - z1t) / (y2t - y1t) + z1t) / dz + nz2
        zb = ((y2 - y1b) * (z2b - z1b) / (y2b - y1b) + z1b) / dz + nz2
        L2 += abs(zt - zb)
        L2 *= abs(y2 - y1) / 2

        for iy in range(ny):
            yc = iy - ny2 + 0.5
            xt = (yc - y1t) * (x2t - x1t) / (y2t - y1t) + x1t + nx2
            xb = (yc - y1b) * (x2b - x1b) / (y2b - y1b) + x1b + nx2
            if xb > xt:
                xt, xb = xb, xt
            if xb >= nx or xt < 0:
                continue
            zt = ((yc - y1t) * (z2t - z1t) / (y2t - y1t) + z1t) / dz + nz2
            zb = ((yc - y1b) * (z2b - z1b) / (y2b - y1b) + z1b) / dz + nz2
            if zb > zt:
                zt, zb = zb, zt
            if zb >= nz or zt < 0:
                continue
            for ix in range(max(0, int(xb)), min(nx, 1 + int(xt))):
                wx = min(ix + 1, xt) - max(ix, xb)
                # wx /= (xt - xb)
                for iz in range(max(0, int(zb)), min(nz, 1 + int(zt))):
                    wz = min(iz + 1, zt) - max(iz, zb)
                    # wz /= (zt - zb)
                    vproj[i] += image[ix, iy, iz] * wx * wz
                    # L += weight
    if L1 * L2 > 0:
        vproj[i] /= L1 * L2


@cuda.jit
def proj_distance_driven_cuda(image, lors, blockdy2, blockdz2, unit_size, center, nb_blocks, vproj):
    i = cuda.grid(1)
    if i >= lors.shape[0]:
        return
    _kernel_proj_3d_distance_driven_cuda(i, image, lors[i, :3], lors[i, 3:],
                                         blockdy2, blockdz2, unit_size, center, nb_blocks, vproj)


@funcclass
class ProjectorDistanceDriven(Projector):
    scanner: PETCylindricalScanner
    device: str = 'gpu'

    def __call__(self, image: Image, lors: LORs):
        nb_blocks = self.scanner.nb_blocks_per_ring
        blockdy2 = self.scanner.blocks.unit_size[1] / 2
        blockdz2 = self.scanner.blocks.unit_size[2] / 2

        if self.device == 'cpu':
            vproj = proj_distance_driven(image.data, lors.data, blockdy2, blockdz2,
                                         image.size / image.shape, image.center, nb_blocks)
        elif self.device == 'gpu':
            # print('projecting in gpu device...')
            blockdim = (256,)
            griddim = (1 + int(len(lors) / blockdim[0]),)
            vproj = np.zeros((len(lors),), dtype = np.float32)
            if not isinstance(image.data, np.ndarray):
                vproj = cuda.device_array((len(lors),), dtype = float32)
            proj_distance_driven_cuda[griddim, blockdim](image.data, lors.data, blockdy2,
                                                         blockdz2, image.size / image.shape,
                                                         image.center, nb_blocks, vproj)
        else:
            raise NotImplementedError
        return Listmode(vproj, lors)



# encoding: utf-8
'''
srfnef.adapter
~~~~~~~~~~~~~~

This module provides adapters to used to match the form srf package.
'''

import json

import h5py
import numpy as np

from .data_types import Block, PETCylindricalScanner, Listmode, LORs


def _block_from_API(path):
    with open(path, 'r') as fin:
        dct = json.load(fin)
        _shape = dct['scanner']['petscanner']['block']['grid']
        _size = dct['scanner']['petscanner']['block']['size']
        _block = Block(_size, _shape)
        return _block


def _scanner_from_API(path):
    _block = _block_from_API(path)
    with open(path, 'r') as fin:
        dct = json.load(fin)
        _inner_radius = dct['scanner']['petscanner']['ring']['inner_radius']
        _outer_radius = dct['scanner']['petscanner']['ring']['outer_radius']
        _nb_rings = dct['scanner']['petscanner']['ring']['nb_rings']
        _nb_blocks_per_ring = dct['scanner']['petscanner']['ring']['nb_blocks_per_ring']
        _gap = dct['scanner']['petscanner']['ring']['gap']
        _scanner = PETCylindricalScanner(_inner_radius, _outer_radius, _nb_rings,
                                         _nb_blocks_per_ring, _gap, _block)
        return _scanner


def load_listmode_from_h5(path, scanner):
    ''' Load old typed listmode data in form of `'.h5`. It stored lors in a `fst` dataset
    and a `snd` dataset. Duplicate lors may appears.

    This function load the h5 file aforementioned and transform it to well-suited `Listmode`
    DataClass tyoe. It will merge repeated lors to valued listmode data.

    '''
    with h5py.File(path, 'r') as fin:
        fst = np.array(fin['listmode_data']['fst'])
        snd = np.array(fin['listmode_data']['snd'])
    return Listmode.from_lors(LORs.from_fst_snd(fst, snd)).compress(scanner)




import numpy as np

from srfnef.data_types import Image, Listmode
from srfnef.func_types import Projector
from srfnef.templating import funcclass, BinaryOps

__all__ = ('UMapProjector',)


@funcclass
class UMapProjector:
    projector: Projector

    def __call__(self, u_map: Image, lors):
        u_map_proj = self.projector(u_map, lors)
        dx = u_map_proj.lors.data[:, 0] - u_map_proj.lors.data[:, 3]
        dy = u_map_proj.lors.data[:, 1] - u_map_proj.lors.data[:, 4]
        dz = u_map_proj.lors.data[:, 2] - u_map_proj.lors.data[:, 5]
        L = np.sqrt(dx ** 2 + dy ** 2 + dz ** 2)
        u_map_proj = u_map_proj * L * L
        return u_map_proj.replace(data = np.exp(-u_map_proj.data))


def corrector(u_map_proj: Listmode, mode = 'outer'):
    if mode == 'outer':
        # return lambda x: mlem(x / u_map_proj)
        return lambda obj: obj @ BinaryOps.Truediv(u_map_proj)
        # def wrapper(obj):
        #     projector2 = obj.projector @ BinaryOps.Truediv(u_map_proj)
        #     return obj.replace(projector = projector2)
        # return wra
    elif mode == 'proj':
        # projector2 = lambda x, lors: u_map_proj * mlem.projector(x, lors)
        def wrapper(obj):
            projector2 = BinaryOps.Mul(u_map_proj) @ obj.projector
            return obj.replace(projector = projector2)

        return wrapper
    elif mode == 'bproj':
        # bprojector2 = lambda listmode: mlem.bprojector(listmode / u_map_proj)
        def wrapper(obj):
            bprojector2 = obj.bprojector @ BinaryOps.Truediv(u_map_proj)
            return obj.replace(bprojector = bprojector2)

        return wrapper
    elif mode == 'proj_bproj':
        # projector2 = lambda x, lors: u_map_proj * mlem.projector(x, lors)
        # bprojector2 = lambda listmode: mlem.bprojector(listmode / u_map_proj)
        def wrapper(obj):
            projector2 = BinaryOps.Mul(u_map_proj) @ obj.projector
            bprojector2 = obj.bprojector @ BinaryOps.Truediv(u_map_proj)
            return obj.replace(projector = projector2, bprojector = bprojector2)

        return wrapper
    else:
        raise AttributeError('mode can only be \' outer\', \' proj\', \' bproj\' or \' '
                             'proj_bproj\'')


from srfnef.templating import TYPE_BIND

TYPE_BIND.update({'UMapProjector': UMapProjector})



# encoding: utf-8
'''
srfnef.templating
~~~~~~~~~~~~~~~~~

This module provides basic data and function types of the whole package,
including `dataclass` for data and `funcclass` for functions.

`DataClass`, as well as `FuncClass`, instances `io` to `.hdf5` files are implemented in `save`
and `load` natual modes.
'''

import json
import operator
import types
from abc import abstractmethod
from copy import copy

import attr
import deepdish as dd
import numpy as np

_tiny = 1e-8

_huge = 1e8

TYPE_BIND = {}


class DataClass:
    '''`DataClass` is suited for storing data objects. It is the most basic class type in
    `srfnef` package. All data and functions would be seriously considered to be defined as a
    Dataclass instance. More specific, only the field `.data` in a `DataClass` would be regarded
    as the only computable part with in an instance. Any other class that defined to contained
    data would a subclass of `DataClass`.
    '''

    def replace(self, **kwargs):
        '''Creates a new object of the same type of instance, replacing fields with values from
        changes.
        '''
        return attr.evolve(self, **kwargs)

    @classmethod
    def fields(cls):
        '''Returns a tuple of field names for this `DataClass` instance.
        '''
        return attr.fields_dict(cls)

    def to_dict(self, recurse = True):
        '''Return a dictionary of fields for this 'DataClass` instance.
        '''
        return attr.asdict(self, recurse)

    def to_tuple(self):
        '''Return a tuple of fields for this 'DataClass` instance.
        '''
        return attr.astuple(self)

    def to_json(self, file_path = ''):
        if not file_path == '' and '.hdf5' in file_path:
            x_saver(file_path, self)
        _dict = {'file_path': file_path, 'classname': self.__class__.__name__}
        _dict.update(self.to_dict(recurse = False))

        for key, val in _dict.items():
            if isinstance(val, np.int64):
                _dict[key] = int(val)
            if isinstance(val, np.ndarray) and not key == 'data':
                _dict[key] = val.tolist()
            if isinstance(val, np.ndarray) and key == 'data':
                _dict[key] = f'numpy.ndarray(shape = {val.shape}, dtype = {val.dtype})'
            if isinstance(val, DataClass):
                _dict[key] = json.loads(val.to_json(file_path + '/' + key))
        return json.dumps(_dict)

    @classmethod
    def from_json(cls, _json):
        _dict = json.loads(_json)
        if not _dict['file_path'] == '' and '.hdf5' in _dict['file_path']:
            return x_loader(_dict['file_path'])
        _class = TYPE_BIND[_dict['classname']]
        _dict.pop('file_path')
        _dict.pop('classname')
        for key, val in _dict.items():
            if isinstance(val, str) and len(val) > 13 and val[:13] == '{"file_path":':
                _dict[key] = DataClass.from_json(val)
        return _class(**_dict)

    def map(self, f):
        '''`Map` applis a function to the field `.data` of this DataClass instance, and return a
        new instance with the same type'''
        return self.replace(data = f(self.data))

    def _map_bin(self, f, other):
        '''`_map_bim` applis a binary function to this instance with another `DataClass`
        instance or np.ndarray or scale values. This function would apply on the field `.data`
        and return a new instance with same type'''

        if isinstance(other, (np.ndarray, list, tuple)) or np.isscalar(other):
            return self.replace(data = f(self.data, other))
        elif isinstance(other, DataClass):
            return self.replace(data = f(self.data, other.data))
        else:
            raise NotImplementedError('{} is not implemented yet.'.format(type(other)))

    @property
    def shape(self):
        ''' return the `shape` of `self.data`.'''
        assert 'data' in self.fields()
        return np.array(self.data.shape)

    def register(self, func):
        obj = self.replace()
        if isinstance(func, list):
            for f in func:
                obj = f(obj)
        elif callable(func):
            obj = func(obj)
        else:
            raise ValueError('func argument must be callable or a list of callables')
        return obj

    @classmethod
    def class_register(cls, func):
        _cls = cls
        if isinstance(func, list):
            for f in func:
                _cls = f(_cls)
        elif callable(func):
            _cls = func(_cls)
        else:
            raise ValueError('func argument must be callable or a list of callables')
        return _cls


def dataclass(cls):
    '''This function is a decorator that is used to add generated special methods to classes, as
    described below.

    The `dataclass()` decorator examines the class to find fields. A field is defined as class
    variable that has a type annotation. The order of the fields in all of the generated methods
    is the order in which they appear in the class definition.
    Be default the instance decorated by `dataclass` is attributed frozen, which means we prefer
    it to be immutable.

    Some basic arithmetic operators are mounted and would apply on `.data` field.
    '''
    base = attr.s(frozen = True, auto_attribs = True, slots = True)(cls)
    cls = types.new_class(base.__name__, (base, DataClass))
    return cls.class_register([unary_ops, binary_ops])


class FuncClass(DataClass):
    '''`FuncClass` is special subclass of `DataClass` which is callable. We prefer all functional
    class in this package would be defined as a subclass of `FuncClass`.

    Be default, the first input argument would be regarded as the key input.
    '''

    @abstractmethod
    def __call__(self, *args, **kwargs):
        '''`__call__` must be defined by any subclass of `FunClass`'''
        pass

    def currying(self, *args, **kwargs):
        ''' `currying` translate this `FuncClass` instance to a single-argument function and
        return it as a anonymous(namely `Curried) `FunClass`. Only the key argument would be the
        only argument in the new instance. The other arguments would be stored in the new `FuncClass`
        instance.
        '''
        return Curried(self, args, kwargs)

    @property
    def is_curried(self):
        '''check this `FuncClass` is `curried`, which means it only accepts one arguments.'''
        import inspect
        return 1 == len(inspect.signature(self.__call__).parameters)

    def __matmul__(self, func):
        ''' return a compostion of two `FuncClass` to a new anonymous(namely `Composited')
        instance. It only support two `is_curried` functions to composite.
        '''
        return Composited(self, func)

    def map(self, f):
        # TODO comments
        if isinstance(f, FuncClass):
            return f @ self
        elif callable(f):
            @funcclass
            class UnaryOp:
                def __call__(self, x):
                    return f(x)

            op = UnaryOp()
            return op @ self
        else:
            raise ValueError('f must be a FuncClass or callable')

    def _map_bin(self, f, other):
        # TODO
        if isinstance(f, FuncClass):
            return (f.currying(other)) @ self
        elif callable(f):
            @funcclass
            class BinaryOp:
                def __call__(self, x):
                    return f(x, other)

            op = BinaryOp()
            return op @ self


def funcclass(cls):
    '''This function is a decorator that is used to add generated special methods to classes, as
    described below.

    The `funcclass()` decorator are used to define function typed class in this package.
    '''
    base = attr.s(frozen = True, auto_attribs = True, slots = True)(cls)
    cls = types.new_class(base.__name__, (base, FuncClass))
    return cls


@funcclass
class Composited(FuncClass):
    f1: FuncClass
    f2: FuncClass

    def __call__(self, *args, **kwargs):
        if isinstance(self.f1, FuncClass):
            assert self.f1.is_curried

        return self.f1(self.f2(*args, **kwargs))


@funcclass
class Curried:
    f: FuncClass
    args: list
    kwargs: dict

    def __call__(self, x):
        return self.f(x, *self.args, **self.kwargs)


class UnaryOps:
    pass


class BinaryOps:
    @funcclass
    class Add:
        other: DataClass

        def __call__(self, obj):
            return obj + self.other

    @funcclass
    class Sub:
        other: DataClass

        def __call__(self, obj):
            return obj - self.other

    @funcclass
    class Mul:
        other: DataClass

        def __call__(self, obj):
            return obj * self.other

    @funcclass
    class Truediv:
        other: DataClass

        def __call__(self, obj):
            return obj / self.other


def unary_ops(obj):
    def abs(obj):
        return obj.map(operator.abs)

    def __neg__(obj):
        return obj.map(operator.neg)

    obj.abs = abs
    obj.__neg__ = __neg__
    return obj


def binary_ops(obj):
    def __eq__(obj: DataClass, other):
        return obj._map_bin(operator.eq, other)

    def __gt__(obj: DataClass, other):
        return obj._map_bin(operator.gt, other)

    def __ge__(obj: DataClass, other):
        return obj._map_bin(operator.ge, other)

    def __lt__(obj: DataClass, other):
        return obj._map_bin(operator.lt, other)

    def __le__(obj: DataClass, other):
        return obj._map_bin(operator.le, other)

    def __add__(obj: DataClass, other):
        return obj._map_bin(operator.add, other)

    def __sub__(obj: DataClass, other):
        return obj._map_bin(operator.sub, other)

    def __mul__(obj: DataClass, other):
        return obj._map_bin(operator.mul, other)

    def __truediv__(obj: DataClass, other):

        def remove_zeros(current, threshold = _tiny):
            if isinstance(current, np.ndarray) or np.isscalar(current):
                _data = copy(current)
            else:
                _data = copy(current.data)
            _data[_data < threshold] = _huge
            return current.replace(data = _data)

        if isinstance(other, DataClass):
            _other = remove_zeros(other)
        else:
            _other = other
        return obj._map_bin(operator.truediv, _other)

    def __floordiv__(obj: DataClass, other):
        return obj._map_bin(operator.floordiv, other)

    def __mod__(obj: DataClass, other):
        return obj._map_bin(operator.mod, other)

    def __pow__(obj: DataClass, other):
        return obj._map_bin(operator.pow, other)

    obj.__eq__ = __eq__
    obj.__gt__ = __gt__
    obj.__ge__ = __ge__
    obj.__lt__ = __lt__
    obj.__le__ = __le__
    obj.__add__ = __add__
    obj.__sub__ = __sub__
    obj.__mul__ = __mul__
    obj.__truediv__ = __truediv__
    obj.__floordiv__ = __floordiv__
    obj.__mod__ = __mod__
    obj.__pow__ = __pow__
    return obj


def _parse_dataclass_to_dict(instance: DataClass):
    dct = {'classname': type(instance).__name__}
    for key, value in instance.to_dict(recurse = False).items():
        if isinstance(value, DataClass):
            dct[key] = _parse_dataclass_to_dict(value)
        else:
            dct[key] = value
    return dct


def _parse_dict_to_dataclass(dct: dict):
    if dct['classname'] not in TYPE_BIND:
        raise NotImplementedError('{} not in TYPE_BIND: {}'.format(dct['classname'],
                                                                   TYPE_BIND.keys()))
    _cls = TYPE_BIND[dct['classname']]
    _dct = {}
    for key, value in dct.items():
        if isinstance(value, dict):
            _dct[key] = _parse_dict_to_dataclass(value)
        elif not key == 'classname' and not key.startswith('_'):
            _dct[key] = value
        else:
            pass
    return _cls(**_dct)


def x_saver(path, instance: DataClass, compression = None):
    """
    Save a DataClass object to an HDF5 file. for more details please visit
    'https://deepdish.readthedocs.io/en/latest/_modules/deepdish/io/hdf5io.html#save'

    This function requires the `deepdish`_ module to be installed.

    Parameters
    ----------
    path : string
        Filename to which the data is saved, must ends with '.hdf5
    data : DataClass or Funcclass
        Data to be saved. This can be any DataClass with fields a Numpy array, a string,
        another DataClass

    compression : string or tuple
        Set compression method, choosing from `blosc`, `zlib`, `lzo`, `bzip2`
        and more (see PyTables documentation). It can also be specified as a
        tuple (e.g. ``('blosc', 5)``), with the latter value specifying the
        level of compression, choosing from 0 (no compression) to 9 (maximum
        compression). The default is None

    See also
    --------
    load
    """
    if path.endswith('.hdf5'):
        dd.io.save(path, _parse_dataclass_to_dict(instance), compression = compression)


def x_loader(path, sel = None, unpack = False):
    """
    Loads an HDF5 saved with `save` and end with '.hdf5'. It will return a DataClass
    object which is well-formed.

    This function requires the `deepdish`_ module to be installed.

    Parameters
    ----------
    path : string
        Filename from which to load the data. must end with '.hdf5'
        path can be in form of path + x_path, for example
        'a.hdf5/b/c'
        it will load the '/b/c' group in a.hdf5
    sel : slice or tuple of slices
        If you specify `group` and the target is a numpy array, then you can
        use this to slice it. This is useful for opening subsets of large HDF5
        files. To compose the selection, you can use `deepdish.aslice`.
    unpack : bool
        If True, a single-entry dictionaries will be unpacked and the value
        will be returned directly. That is, if you save ``dict(a=100)``, only
        ``100`` will be loaded.

    Returns
    -------
    data : DataClass
        It will visit TYPE_BIND to find the bind bewteen a string and a specific
        DataClass class.

    See also
    --------
    save
    TYPE_BIND
    """
    if not '.hdf5' in path:
        return
    _path, x_path = path.split('.hdf5')
    _path += '.hdf5'
    if x_path == '':
        x_path = None

    return _parse_dict_to_dataclass(dd.io.load(_path, x_path, sel, unpack))


def save(*args, **kwargs):
    '''same as `x_loader`

    See also
    --------
    x_loader, load
    '''
    x_saver(*args, **kwargs)


def load(*args, **kwargs):
    '''same as `x_saver`

    See also
    --------
    x_saver, load
    '''
    return x_loader(*args, **kwargs)


@funcclass
class Saver:
    '''This type of `FuncClass` defined the saver during iteration.'''
    path: str
    interval: int

    def __call__(self, ind, x):
        if self.interval == 0:
            return
        if ind % self.interval == 0:
            x_saver(self.path.replace('?', str(ind)), x)


def saver(path, interval = 0):
    def wrapper(obj):
        _saver = Saver(path, interval)
        return obj.replace(saver = _saver)

    return wrapper


@funcclass
class Noop:
    '''Do nothing'''

    def __call__(self, *args, **kwargs):
        return


noop = Noop()


@funcclass
class Indentity:
    '''Indentity'''

    def __call__(self, *args, **kwargs):
        return args, kwargs


TYPE_BIND.update({'DataClass': DataClass, 'FuncClass': FuncClass, 'Saver': Saver, 'Noop': Noop})
TYPE_BIND.update({'Add': BinaryOps.Add, 'Sub': BinaryOps.Sub, 'Mul': BinaryOps.Mul, 'Truediv':
    BinaryOps.Truediv})

TYPE_BIND.update({'Composited': Composited, 'Curried': Curried})



from setuptools import setup, find_packages

setup(name = 'srfnef',
      version = '0.5.1.2',
      py_modules = ['srfnef'],
      description = 'Scalable Reconstruction Framework -- Not Enough Functions',
      author = 'Minghao Guo',
      author_email = 'mh.guo0111@gmail.com',
      license = 'Apache',
      # packages = ['srfnef'],
      packages = find_packages(),
      install_requires = [
          'scipy',
          'matplotlib',
          'typing',
          'h5py',
          'click',
          'numpy',
          'tqdm',
          'numba',
          'deepdish',
          'Click',
      ],
      zip_safe = False,
      entry_points = '''
        [console_scripts]
        srfnef=srfnef.app.cli:cli
      ''',
      )
